import matplotlib.pyplot as plot
from math import inf, nan
from copy import copy, deepcopy
from numpy import power, sqrt, isscalar, float64, diff


class Zerodict(dict):
    def __init__(self, zero=0):
        self.zero = zero
    def __getitem__(self, k):
        return self.setdefault(k, deepcopy(self.zero))
    
class Dist:
    def __init__(self):
        self.moment0 = 0
        self.moment1 = 0
        self.moment2 = 0
        self.min = inf
        self.max = -inf
        self.samples = [] # only for violin, not collected for mass data
    def mean(self):
        return self.moment1/self.moment0
    def std(self):
        return sqrt(self.moment2/self.moment0 - (self.moment1/self.moment0)**2)
    def sum():
        return moment1
    def count():
        return moment0
    def fill(self, values):
        values = float64(values)
        self.moment0 += values.size
        self.moment1 += values.sum()
        self.moment2 += power(values,2).sum()
        self.min = min(values.min(), self.min)
        self.max = max(values.max(), self.max)
        if isscalar(values):
          self.samples.append(values)

    def combine(self, other):
            self.moment0 += other.moment0
            self.moment1 += other.moment1
            self.moment2 += other.moment2
            self.min = min(other.min, self.min)
            self.max = max(other.max, self.max)
            self.samples += other.samples
    def __repr__(self):
        return "Dist(mean = %f, err = %f, n = %d, min = %f, max = %f)" % (self.mean(), self.std(), self.moment0, self.min, self.max)
    
class Profile:
    def __init__(self, name=""):
        self.data = Zerodict(Dist())
        self.axislabels = []
        self.ylabel = name
        
    def bin(self, x):
        return x
        
    def fill(self, *args):
        idx = tuple(self.bin(x) for x in args[1:])
        self.data[idx].fill(args[0])
    
    def plot(self, xaxis=1, fig=None, colors="bgrcmyk", markers="osv^*<>", violin=False):
        if not fig:
            fig = plot.figure(figsize=(14,11))
        ax = fig.gca()
        ax.grid()

        c_map = {None: 0}
        m_map = {None: 0}
        def style(tup):
            it = (item for item in tup if item is not None)
            first = next(it, ())
            tail = tuple(it)
            if len(tail) == 0: tail = first
            if not tail in m_map:
                m_map[tail] = markers[m_map[None] % len(markers)]
                m_map[None] += 1
            if not first in c_map:
                c_map[first] = colors[c_map[None] % len(colors)]
                c_map[None] += 1
            return "%s%s-" % (c_map[first], m_map[tail])
        
        if not isinstance(xaxis, int):
            xaxis = self.axislabels.index(xaxis)
        lines = Zerodict([])
        for t in self.data:
           l = list(t)
           x = l[xaxis]
           l[xaxis] = None
           lines[tuple(l)].append(x)
        lineidx = list(lines.keys())
        lineidx.sort()
        for l in lineidx:
            xs = lines[l]
            sty = style(l)
            xs.sort()
            ys = []
            errs = []
            samples = []
            for x in xs:
                i = list(l)
                i[xaxis] = x
                i = tuple(i)
                ys.append(self.data[i].mean())
                errs.append(self.data[i].std())
                samples.append(self.data[i].samples)
            if isinstance(xs[0], str):
                ax.set_xticks(range(len(xs)))
                ax.set_xticklabels(xs, rotation='vertical')
                xs = range(len(xs))
            s = " ".join([self.axislabels[i]+"="+str(l[i]) for i in range(len(l)) if i != xaxis and l[i] != None])
            ax.errorbar(xs, ys, yerr=errs, fmt=sty, label=s)
            if violin:
                res = ax.violinplot(samples, xs, showmeans=False, showextrema=False, showmedians=False, widths=diff(xs).mean()*0.2)

        ax.xaxis.set_label_text(self.axislabels[xaxis])
        ax.yaxis.set_label_text(self.ylabel)
        ax.legend(loc="best")
            
        return fig
    
    def filter(self, key, value = None, pred = None):
        if pred == None:
            pred = lambda x: x == value
        out = copy(self)
        out.data = Zerodict(Dist())
        idx = self.axislabels.index(key)
        for k in self.data:
            if pred(k[idx]):
                out.data[k] = self.data[k]
        return out
                
    def group(self, *keys):
        out = copy(self)
        out.data = Zerodict(Dist())
        idxs = [self.axislabels.index(key) for key in keys]
        for k in self.data:
            l = list(k)
            for idx in idxs:
                l[idx] = None
            out.data[tuple(l)].combine(self.data[k])
        return out
    
    def map(self, func, name=None):
        out = copy(self)
        if name != None:
            out.ylabel = name
        out.data = Zerodict(Dist())
        for k in self.data:
            out.data[k].fill(func(self.data[k], k))
        return out
    
    def __repr__(self):
        return """
Profile(ylabel = %s,
        axislabels = %s,
        data = {
%s})
               """ % (repr(self.ylabel), repr(self.axislabels), "\n".join("%s: %s" % (repr(k), repr(v)) for k, v in self.data.items()))

if __name__ == '__main__':  
    from random import normalvariate as normal
    p = Profile("Normal distibutions")
    p.axislabels = ["i", "mu", "sigma"]
    for mu in range(0, 3):
        for sigma in range(0, 3):
            for i in range(1, 100):
                p.fill(normal(mu, sigma), i, mu, sigma)
    print(repr(p.group("i")))
    print(repr(p.group("i").filter("mu", 1)))
    p.group("i").plot(xaxis="sigma").savefig("test.pdf")
    
