#!/usr/bin/env python3
import random
import numpy as np
from collections import defaultdict
from simplepmap import pmap

from countpaths import *
from count_collisions import *

# convention: g[s,t] = x -- route from s to t via x
def directroutes(g):
    out = np.zeros(g.shape, np.int64)
    for s in range(g.shape[0]):
        for t in range(g.shape[1]):
            out[s,t] = t if g[s,t] != 0 else -1
    return out

# Make a ECMP MIN routing table for g, nexthops are set of node id
def make_shortest(g, maxlen):
    empty = np.frompyfunc(lambda x: set(),1,1)
    toset = np.frompyfunc(lambda x: set([x]) if x >= 0 else set(),1,1)
    
    def left(a, b):
        out = empty(a)
        for i in range(len(out)):
            if len(b[i]) != 0:
                out[i] = set(a[i])
        return out
    def union(a):
        # overloaded for set, so this works...
        return np.bitwise_or.reduce(a)
    
    def mmm(outer, inner, a, b):
        out = np.zeros_like(a)
        for i in range(a.shape[0]):
            for j in range(a.shape[1]):
                out[i,j] = outer(inner(a[i,:], b[:,j]))
        return out
    
    out = empty(g)
    
    acc = toset(directroutes(g))
    g = toset(directroutes(g))
    
    for i in range(maxlen):
        newshortest = (out == set()) & (acc != set())
        out[newshortest] = acc[newshortest]
        acc = mmm(union, left, acc, g)

    return out

def bfs_shortest_path(s,t, g):
    if s == t: return []
    q = [s]
    marker = -np.ones(g.shape[0], dtype = np.int64)
    while len(q) > 0:
        cur = q.pop()
        out = np.nonzero((g[cur,:] == 1) & (marker == -1))[0] # 1-el tuple...
        if t in out:
            path = [(cur, t)]
            while cur != s:
                path = [(marker[cur], cur)] + path
                cur = marker[cur]
            return path
        marker[out] = cur
        newq = list(out)
        random.shuffle(newq) # prevent bias in k-shortest path if k is smaller than path multplicity...
        q = newq + q
    return None

def restricted_bfs(s,t, g, limit = 100, shape = "aaaaa", groupsize = 1):
    if random.randint(0, 999) == 0: print(".")
    if s == t: return [[]]
    # any, local, global. Uppercase for "may not terminate here"
    filters = {'a': lambda s, t: True,
               'l': lambda s, t: s//groupsize == t//groupsize,
               'g': lambda s, t: s//groupsize != t//groupsize}
    
    q = [[ [] ]] # list of lists of paths, initialized with empty path

    paths = []
    maxdist = len(shape)
    while len(q) > 0:
        curpath = q[0].pop()
        if len(q[0]) == 0: del q[0] # nested q-of-q's
        curdist = len(curpath)
        cur = curpath[-1][1] if curdist > 0 else s # last node in path, but may be empty
        if curdist >= maxdist: break # we checked for hits while expanding, so we have all hits already
        pred = filters[shape[curdist].lower()]
    
        fringe = np.nonzero((g[cur,:] == 1))[0] # 1-el tuple...
    
        newq = []
        for node in fringe:
            if not pred(cur, node): continue
            path = curpath + [(cur, node)]
            if node == t and shape[curdist].islower(): # termination here is allowed
                maxdist = curdist+1
                paths.append(path)
                if len(paths) == limit: break # TODO: bias here.
            else:
                newq.append(path)
        
        random.shuffle(newq) # prevent bias in k-shortest path if k is smaller than path multplicity...
        q.append(newq)
    return paths

def make_nondirect_paths(s,t, g):
    # shortest possible paths longer than one hop (and shorter than 6)
    return restricted_bfs(s,t, g, shape = "Aaaaaa")

def make_lgl_paths(s,t, g, p):
    # "lgl" covers group local, but not leaving the group immediatly
    return (restricted_bfs(s,t, g, shape = "lgl", groupsize = 2*p) +
            restricted_bfs(s,t, g, shape = "gl", groupsize = 2*p))
           
def make_lglgl_paths(s,t, g, p):
    # since lgl already reaches all nodes, and we still restrict to shortest,
    # we need to force any combination of detour here
    return (restricted_bfs(s,t, g, shape = "lGLGl", groupsize = 2*p) +
            restricted_bfs(s,t, g, shape = "GLGl", groupsize = 2*p) +
            restricted_bfs(s,t, g, shape = "lGGl", groupsize = 2*p))

def make_kshortest_paths(s,t, g, K):
    if s == t: return [[]] * K # s==t is not legal here, "loop free"
    g = np.array(g) # clone, we change it

    # using WP notation: https://en.wikipedia.org/wiki/Yen%27s_algorithm
    A = [bfs_shortest_path(s,t, g)]
    B = []
    
    for k in range(1, K):
        for i in range(0, len(A[k-1])):
            spurNode = A[k-1][i][0] # all nodes in path but t
            rootPath = A[k-1][0:i] # segments up to spurNode
            for p in A:
                if rootPath == p[0:i]:
                    for s,t in p: g[s,t] = 2
            for seg in rootPath:
                g[seg[0],:] *= 2
                g[:,seg[0]] *= 2
            spurPath = bfs_shortest_path(spurNode, t, g)
            if spurPath: # there might not be a path
                B.append(rootPath + spurPath)
            g[g > 1] = 1 # restore marked edges
            
        if len(B) == 0: break
        B.sort(key=len, reverse = True) # shortest in the end
        A.append(B.pop()) # pop from back
    return A

# given a ECMP table, recursively enumerate all ECMP paths
# a path/route is a list of (s,t) pairs
def make_ecmp_paths(s,t, table):
    if s == t: return [[]] # one path, we are there
    out = [] # collect paths
    for x in table[s,t]:
        if x == s: continue
        out += [[(s,x)] + rest for rest in make_ecmp_paths(x, t, table)]
    return out

def make_valiant_paths(s,t, g, limit):
    if s == t: return [[]]
    nodes = range(g.shape[0])
    hops = random.sample(nodes, limit) if limit < g.shape[0] else list(nodes)
    return [bfs_shortest_path(s,h, g) + bfs_shortest_path(h,t, g) for h in hops]

def make_layers(g, number, amount):
    def sparsify(g, amount):
        g = np.array(g)
        total = int(g.sum())
        remaining = int(total*amount)
        values = np.array([1]*remaining + [0]*(total-remaining))
        random.shuffle(values)
        g[g > 0] = values
        return g
    return [np.array(g)] + [sparsify(g, amount) for _ in range(number-1)]

def make_layered_paths(s,t, layers):
    return list(filter(lambda x: x is not None, (bfs_shortest_path(s, t, l) for l in layers)))

def make_spain_paths(s,t, g, number):
    g = np.array(g)
    paths = []
    for i in range(number):
        p = bfs_shortest_path(s, t, g)
        if p is None: break
        for x,y in p:
            g[x,y] = 0
        paths.append(p)
    return paths

# make a flat list of ((s,t), [route]) potential flow routes, given permutations
# and a routing function.
def routeflows(workload, p, routefunc):
    print("Routing %d flows..." % (len(workload)*len(workload[0])))
    out = []
    for pi in workload:
        out += pmap(lambda st: [(st, route) for route in routefunc(st[0]//p,st[1]//p)], list(enumerate(pi)))
    if len(list(filter(lambda x: len(x) == 0, out))) > 0:
        print("Warning: unroutable flows")
    return sum(out, []) # flatten

OVERCOMMIT=1
# no real packing here. Just try to add flows in random order until all flows are
# mapped or the network is congested.
def packflows(flows, g):
    print("Packing %d routes..." % len(flows))
    g = np.array(g) # note that we don't actually need g... ones would be fine.
    g[g > 0] = OVERCOMMIT
    f = defaultdict(lambda: False) # where we have flows (endpoint s,t) 
    mapped = 0
    random.shuffle(flows) # randomize workload AND route order
    flows.sort(key = lambda f: len(f[1])) # but prefer short routes 
    for flow, route in flows:
        if not f[flow]:
            if all(g[s,t] for s,t in route):
                f[flow] = True
                for s,t in route:
                    g[s,t] -= 1
                mapped += 1
    return mapped, g.sum()

#g = np.array([[0,1,1,0],
              #[1,0,0,1],
              #[1,0,0,1],
              #[0,1,1,0]])
#table = make_shortest(g,3)
#make_ecmp_paths(0, 3, table)
#bfs_shortest_path(0, 3, g)
#make_kshortest_paths(0, 1, g, 3)
#flows = routeflows(randstencil(4,2), 1, lambda s,t: make_ecmp_paths(s,t, table))
#routeflows(randstencil(4,2), 1, lambda s,t: make_kshortest_paths(s,t, g, 2))
#packflows(flows, g)

if __name__ == "__main__":
    from results import Results
    r = Results("flowpacker.db")
    for topo, (files, samp) in graphnames.items():
        for i in range(1,samp+1):
            for name in files:
                if i != 1:
                    n = name[:-1] + str(i)
                else:
                    n = name
                print(n)
                g = read_graph(n)
                h = np.array(g)
                # p on _edge switches_
                p = round(g.radix / designd[topo] * g.vertices / g.edge)
                #p = 1
                #rt = make_shortest(np.array(g), 5)
                layers_0_3_10 = make_layers(g, 10, 0.3)
                layers_0_5_10 = make_layers(g, 10, 0.5)
                layers_0_7_10 = make_layers(g, 10, 0.7)
                layers_shortest_10 = [np.array(g)] * 10
                routers = {"1-s'est":   lambda s,t: make_kshortest_paths(s,t, g, 1),
                           "10-s'est":  lambda s,t: make_kshortest_paths(s,t, g, 10),
                           "100-s'est": lambda s,t: make_kshortest_paths(s,t, g, 100),
                           #"ECMP":      lambda s,t: make_ecmp_paths(s,t, rt),
                           "MIN":       lambda s,t: restricted_bfs(s,t, h),
                           #"1-VAL":     lambda s,t: make_valiant_paths(s,t, h, 1),
                           "10-VAL":    lambda s,t: make_valiant_paths(s,t, h, 10),
                           #"min2":      lambda s,t: make_nondirect_paths(s,t, h),
                           #"lgl":       lambda s,t: make_lgl_paths(s,t, h, p),
                           #"lglgl":     lambda s,t: make_lglgl_paths(s,t, h, p),
                           "10-L-0.3":  lambda s,t: make_layered_paths(s,t, layers_0_3_10),
                           "10-L-0.5":  lambda s,t: make_layered_paths(s,t, layers_0_5_10),
                           "10-L-0.7":  lambda s,t: make_layered_paths(s,t, layers_0_7_10),
                           "10-L-1.0":  lambda s,t: make_layered_paths(s,t, layers_shortest_10),
                           #"1-dis't":   lambda s,t: make_spain_paths(s,t, h, 1),
                           #"3-dis't":   lambda s,t: make_spain_paths(s,t, h, 3),
                           "10-dis't":  lambda s,t: make_spain_paths(s,t, h, 10),
                           }
                workload = randperm(g.edge * p)
                for rname, rfunc in routers.items():
                    print(rname)
                    collect = r.collector(count=Results.Int, unused=Results.Int, topo=topo, n_r=g.vertices, radix=g.radix, n_e=g.edge*p, p=p, designd=designd[topo], sample=i, workload="randperm", router=rname, overcommit=OVERCOMMIT)
                    flows = map(lambda w: routeflows(w, p, rfunc), [workload] * 3)
                    res = pmap(lambda f: packflows(f, g), list(flows)*3)
                    for count, unused in res:
                        collect(count=count, unused=unused)
                    r.commit()
    r.close()






