#!/usr/bin/env python3
import sqlite3
import inspect
import subprocess

def growtable(conn, table, newcolumns):
    c = conn.execute("SELECT * FROM %s LIMIT 1;" % table)
    columns = [d[0].lower() for d in c.description]
    for k, t in newcolumns.items():
        if k.lower() not in columns:
            q = "ALTER TABLE %s ADD COLUMN %s %s;" % (table, k, t.typename)
            print(q)
            conn.execute(q)
            
def initdb(dbfile):
    conn = sqlite3.connect(dbfile)
    load_sqlite_ext(conn)
    conn.execute("CREATE TABLE IF NOT EXISTS runs(runid INTEGER PRIMARY KEY AUTOINCREMENT, timestamp text, githash text, file text);")
    conn.execute("CREATE TABLE IF NOT EXISTS datapoints(runid INTEGER REFERENCES runs(runid));")
    conn.execute("CREATE INDEX IF NOT EXISTS runtodata ON datapoints(runid);")
    return conn

def load_sqlite_ext(conn):
    from pathlib import Path
    from os.path import realpath
    conn.enable_load_extension(True)
    module = Path(realpath(__file__)).parent / "percentile.so"
    if not module.exists():
        subprocess.call(['gcc', '-fPIC', '--shared', '-o', str(module), str(module.parent/'percentile.c')])
    conn.load_extension(str(module))
    module = Path(realpath(__file__)).parent / "extension-functions.so"
    if not module.exists():
        subprocess.call(['gcc', '-fPIC', '--shared', '-o', str(module), str(module.parent/'extension-functions.c')])
    conn.load_extension(str(module))

def make_commit():
    # fails if there is nothing to commit, which is fine.
    tempcommit = subprocess.call('git commit -a -m "Automatic commit for run on `date`" 1>&2', shell=True)
    res = subprocess.call('git tag -f test-`date "+%Y-%m-%d_%H_%M_%S"` 1>&2', shell=True)
    #assert res == 0, "Could not tag software in git"
    githash = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode().strip()
    if tempcommit == 0: # success
        res = subprocess.call('git reset HEAD^ 1>&2', shell=True)
        assert res == 0, "Could not reset state in git"
    return githash

class Results:
    # unique markers for to-be-added columns
    # note that SQLite is always dynamically typed, and all columns can take any values
    # there will be some limited conversions when a type is given.
    class Type:
        def __init__(self, name, convert):
            self.typename = name;
            self.convert = convert
    Int = Type('INTEGER', int)
    Real = Type('REAL', float)
    Text = Type('TEXT', str)
    Any = Type('', lambda x: x)
    types = {Int, Real, Text, Any}
    
    def __init__(self, datafile = 'results.db', githash = None):
        self.conn = initdb(datafile)
        
        self.githash = githash if githash else make_commit()
            
        # to make sure timestamp is fixed for this run
        self.timestamp = next(self.conn.execute("SELECT datetime('now');"))[0]
        
    def collector(self, **parameters):
        fixedparameters = {k: v for k, v in parameters.items() if v not in Results.types}
        fixedparameters['timestamp'] = self.timestamp
        fixedparameters['githash'] = self.githash
        fixedparameters['file'] = inspect.stack()[1].filename
        varparameters = {k: v for k, v in parameters.items() if v in Results.types}
        varparameters['runid'] = Results.Int

        keys = sorted(fixedparameters.keys())
        varkeys = sorted(varparameters.keys())
        varconvert = [varparameters[k].convert for k in varkeys]
        growtable(self.conn, 'runs', {k: Results.Any for k in fixedparameters})         
        growtable(self.conn, 'datapoints', varparameters)         
        
        self.conn.execute("INSERT INTO runs(%s) VALUES (%s);" 
                          % (", ".join(keys), ", ".join(["?"]*len(keys))),
                          [fixedparameters[k] for k in keys])
        
        runid = next(self.conn.execute("SELECT last_insert_rowid();"))[0]
        
        insert = ("INSERT INTO datapoints(%s) VALUES (%s);" 
                          % (", ".join(varkeys), ", ".join(["?"]*len(varkeys))))
        
        def collect(**kws):
            kws['runid'] = runid
            assert len(kws) == len(varkeys)
            values = [conv(kws[k]) for k, conv in zip(varkeys, varconvert)]
            self.conn.execute(insert, values)
        
        return collect;
    
    def close(self):
        self.conn.commit()
        self.conn.close()
        
    def commit(self):
        self.conn.commit()
        
def mergeresults(datafile, addedfile):
    assert datafile != addedfile
    conn = initdb(datafile)
    conn.execute("ATTACH DATABASE ? AS new;", (addedfile,))
    
    c = conn.execute("SELECT * FROM new.runs LIMIT 1;")
    columns =  {d[0].lower(): Results.Any for d in c.description}
    growtable(conn, "runs", columns)
    newruncols = ", ".join(k for k in columns if k != "runid")
    
    c = conn.execute("SELECT * FROM new.datapoints LIMIT 1;")
    columns =  {d[0].lower(): Results.Any for d in c.description}
    growtable(conn, "datapoints", columns)
    newdatacols = ", ".join(k for k in columns if k != "runid")
    
    c = conn.execute("SELECT * FROM runs LIMIT 1;")
    runcols = ", ".join(d[0].lower() for d in c.description if d[0] != "runid")
    
    conn.execute("CREATE TEMP TABLE newids AS SELECT runid FROM new.runs WHERE (%s) NOT IN (SELECT DISTINCT %s FROM runs);" % (newruncols, newruncols))
    conn.execute("INSERT INTO runs(%s) SELECT %s FROM new.runs WHERE runid IN newids;" % (newruncols, newruncols))
    
    conn.execute("CREATE TEMP VIEW newruns(%s,  newid) AS SELECT %s,  rowid FROM runs;" % (runcols, runcols))
    conn.execute("CREATE TEMP VIEW idmap(oldid, newid) AS SELECT runid, newid FROM newruns NATURAL JOIN new.runs;")
    conn.execute("""INSERT INTO datapoints(runid,  %s) SELECT newid,  %s 
                    FROM new.datapoints INNER JOIN idmap ON runid = oldid 
                    WHERE runid in newids;""" % (newdatacols, newdatacols))
    conn.commit()
    conn.close()

def plotdata(datafile, sql = None, runsql = None, datasql = None, runwhere = None, where = None, select = "count(*)", group = False, ignore = None):
    conn = initdb(datafile)
    
    if not sql:
        if not runsql:
            c = conn.execute("SELECT * FROM runs LIMIT 1;")
            cols = ", ".join(d[0] for d in c.description if d[0] not in ["runid", "githash", "timestamp"])
            if runwhere:
                clause = "WHERE %s" % runwhere
            else:
                clause = ""
            runsql = "SELECT max(timestamp), * FROM runs %s GROUP BY %s" % (clause, cols)
            
        if not datasql:
            if group:
                c = conn.execute("SELECT * FROM runs LIMIT 1;")
                rcols = [d[0] for d in c.description if d[0]]
                c = conn.execute("SELECT * FROM datapoints LIMIT 1;")
                dcols = [d[0] for d in c.description if d[0]]
                if not ignore:
                    ignore = dcols + ["runid"]
                else:
                    ignore = ignore + ["runid"]
                cols = [c for c in dcols + rcols if c not in ignore]
                clause = "GROUP BY %s" % ", ".join(cols)
            else:
                clause = ""
                
            datasql = "SELECT %s FROM data %s" % (select, clause)
        
        if where != None:
            whereclause = "WHERE %s" % where
        else:
            whereclause = ""
            
        sql = """
            WITH current AS (
                %s
            ), data AS (
                SELECT * 
                FROM datapoints 
                INNER JOIN current ON current.runid = datapoints.runid
                %s
            )
            %s""" % (runsql, whereclause, datasql)
                 
    return conn, sql

if __name__ == '__main__':
    from sys import stdin
    def record(datafile, parameter, variable, githash):
        typemap = {"any": Results.Any, "int": Results.Int, "text": Results.Text, "real": Results.Real}
        args = {k: v for k,v in parameter}
        args.update({k: typemap[t.lower()] for k,t in variable})
        r = Results(datafile, githash)
        collector = r.collector(**args)
        for l in stdin.readlines():
            if l.isspace(): continue
            collector(**{k[0]:v for k, v in zip(variable, l.split())})
        r.close()
    
    def execute(datafile):
        code = stdin.read()
        conn = sqlite3.connect(datafile)
        conn.executescript(code)
        
    def show(explain=False, limit=0, **kwargs):
        conn, sql = plotdata(**kwargs)
        if limit > 0:
            sql = "%s LIMIT %i" % (sql, limit)
        print(sql)
        def prettyprint(c):
            if not c.description: print("No data."); return
            print("\t".join(d[0] for d in c.description))
            for row in c:
                print("\t".join(str(v) for v in row))
        if explain:
            prettyprint(conn.execute("EXPLAIN " + sql))
            prettyprint(conn.execute("EXPLAIN QUERY PLAN " + sql))
        else:
            prettyprint(conn.execute(sql))
        conn.close()
        
    def listruns(**kwargs):
        show(datasql="SELECT * FROM runs", **kwargs)
        
    def merge(datafile, addedfiles):
        for f in addedfiles:
            mergeresults(datafile, f)
            
    def ggplot(**kwargs):
        import results_ggplot
        results_ggplot.ggplot(**kwargs)
        
    def ggplot2(**kwargs):
        import results_ggplot2
        results_ggplot2.ggplot(**kwargs)
    def commit(**kwargs):
        print(make_commit())

    import argparse
    parser = argparse.ArgumentParser(prog='results.py')
    parser.add_argument('-f', '--datafile', help='SQLite file to operate on', default="results.db")
    subparsers = parser.add_subparsers(help='operation')
    
    parser_execute = subparsers.add_parser('execute', help='Execute SQL code from stdin.')
    parser_execute.set_defaults(func=execute)
    
    parser_commit = subparsers.add_parser('commit', help='Create and tag an automatic commit and print the hash.')
    parser_commit.set_defaults(func=commit)
    
    parser_record = subparsers.add_parser('record', help='Record results from stdin.')
    parser_record.add_argument('-p', '--parameter', nargs=2, metavar=("PARAMETER", "VALUE"), help='Set per-run parameter to value.', action='append')
    parser_record.add_argument('-v', '--variable', nargs=2, metavar=("VARIABLE", "TYPE"), help='Declare per-datapoint variable with given type [Any, Int, Real, Text]. The values shall appear whitespace-separated in the same order on stdin, one datapoint per line.', action='append')
    parser_record.add_argument('--githash', help='Use this githash instead of creating a commit and tag.', default=None)
    parser_record.set_defaults(func=record)
    
    parser_list = subparsers.add_parser('list', help='List runs in datafile.')
    parser_list.add_argument('--limit', help='Limit to this number of results (0 for no limit).', type=int, default=100)
    parser_list.set_defaults(func=listruns)
   
    parser_merge = subparsers.add_parser('merge', help='Merge results from another file into the current DB.')
    parser_merge.add_argument('addedfiles', nargs='*', help='the files to be merged')
    parser_merge.set_defaults(func=merge)
   
    parser_show = subparsers.add_parser('show', help='Execute a SQL query and show the results.')
    parser_show.add_argument('--explain', help='Only show EXPLAIN output.', action='store_true')
    parser_show.add_argument('--limit', help='Limit to this number of results (0 for no limit).', type=int, default=100)
    parser_show.set_defaults(func=show)
    
    parser_ggplot = subparsers.add_parser('ggplot', help='Execute a SQL query and plot results using python-ggplot. Uses matplotlib, pandas and the ggplot package.')
    parser_ggplot.set_defaults(func=ggplot)
    
    parser_ggplot2 = subparsers.add_parser('ggplot2', help='Execute a SQL query and plot results using R ggplot2. Requires sqldf and ggplot2 for R.')
    parser_ggplot2.set_defaults(func=ggplot2)
    parser_ggplot2.add_argument('--manual', help='Do not run R, just generate script and data.', action='store_true')
    
    for sub in [parser_ggplot, parser_ggplot2]:
        sub.add_argument('-o', '--outfile', help='Output plot file name.', default="plot.pdf")
        sub.add_argument('--size', help='Plot size (e.g. 10x12, inches).', default="12x10")
    
    for sub in [parser_show, parser_ggplot, parser_ggplot2]:
        sub.add_argument('--sql', help='Use this complete SQL query.', default=None)
        sub.add_argument('--runsql', help='Use this SQL SELECT to select runs.', default=None)
        sub.add_argument('--datasql', help='Use this SQL SELECT to select datapoints.', default=None)
        sub.add_argument('-r', '--runwhere', help='Use this SQL WHERE clause to select runs.', default=None)
        sub.add_argument('-w', '--where', help='Use this SQL WHERE clause to select datapoints.', default=None)
        sub.add_argument('-s', '--select', help='Output these columns.', default='*, count(*)')
        sub.add_argument('-g', '--group', help='Group the result by run.', action='store_true')
        sub.add_argument('-i', '--ignore', help='Group by all columns except this. Can be used multiple times.', action='append')
    
    args = parser.parse_args()
    func = args.func
    args.__dict__
    del args.func
    func(**args.__dict__)
   
    
    
