# Plot function using R ggplot2.
from results import plotdata
import subprocess
from tempfile import NamedTemporaryFile

PRINTTHEME = True

def ggplot(outfile, size, manual, **kwargs):
    w, h = [float(x) for x in size.split("x", 1)]

    conn, sql = plotdata(**kwargs)
    header = sql + " LIMIT 0"
    c = conn.execute(header)
    cols  = [d[0] for d in c.description]
    paras = [c.split(';', 1) for c in cols]
    paras = [(c[0].split(' '), c[1] if len(c) > 1 else c[0], name) for c, name in zip(paras, cols)]
    rname = {c : 'C%d' % i for i, c in enumerate(cols)}

    propertheme = """
library(sqldf)
library(ggplot2)
library(scales)

theme_Publication <- function(base_size=14, base_family="sans") {
      library(grid)
      library(ggthemes)
      (theme_foundation(base_size=base_size, base_family=base_family)
       + theme(plot.title = element_text(face = "bold",
                                         size = rel(1.2), hjust = 0.5),
               text = element_text(),
               panel.background = element_rect(colour = NA),
               plot.background = element_rect(colour = NA),
               panel.border = element_rect(colour = NA),
               axis.title = element_text(face = "bold",size = rel(0.9)),
               axis.title.y = element_text(angle=90,vjust =2),
               axis.title.x = element_text(vjust = -0.2),
               axis.text = element_text(),
               axis.line = element_line(colour="black"),
               axis.ticks = element_line(),
               panel.grid.major = element_line(colour="#c0c0c0"),
               panel.grid.minor =  element_line(colour="#c0c0c0"),
               legend.key = element_rect(colour = NA),
               legend.position = "bottom",
               legend.direction = "horizontal",
               legend.key.size= unit(0.6, "cm"),
               legend.margin = unit(0, "cm"),
               legend.title = element_text(face="italic"),
               plot.margin=unit(c(3,2,2,2),"mm"),
               strip.background=element_rect(colour="#f0f0f0",fill="#f0f0f0"),
               strip.text = element_text(face="bold")
          ))

}

scale_fill_Publication <- function(...){
      library(scales)
      discrete_scale("fill","Publication",manual_pal(values = c("#7fc97f","#ef3b2c","#662506","#386cb0","#fdb462","#a6cee3","#fb9a99","#984ea3","#ffff33")), ...)

}

scale_colour_Publication <- function(...){
      discrete_scale("colour","Publication",manual_pal(values = c("#4e974f","#386cb0","#ef3b2c","#fdb462","#a6cee3","#fb9a99","#984ea3","#ffff33")), ...)

}

scientific_10 <- function(x) {
    xout <- gsub("1e", "10^{", format(x, scientific = TRUE),fixed=TRUE)
    xout <- gsub("NA", "{NA", xout,fixed=TRUE)
    xout <- gsub("{-0", "{-", xout,fixed=TRUE)
    xout <- gsub("{+", "{", xout,fixed=TRUE)
    xout <- gsub("{0", "{", xout,fixed=TRUE)
    xout <- paste(xout,"}",sep="")
    return(parse(text=xout))
}
   """

    # DB name is added later, once we have the temp db. 
    rcode = [propertheme, 
             "labels <- sqldf('SELECT * FROM labels', dbname='%s')",
             "data   <- sqldf('SELECT * FROM data', dbname='%s')",
             "plot   <- ("] # to allow arbitrary linebreaks
    
    aes = dict()
    labs = dict()
    
    def matchrule(key, action):
        for kw, lab, col in paras:
            if all(k in kw for k in key): action(kw, lab, col)
            
    def setcol(key):
        def setit(kw, lab, col):
            assert key not in aes, "Multiple values for %s (%s and %s)" % (key, aes[key], col)
            aes[key] = rname[col]
            labs[key] = "labels[1,'%s']" % rname[col]
            if lab == "": labs[key] = "''" # to hide the guide later
        return setit
    
    def hist(kw, lab, name):
        count = ("alpha" if "fill" in aes else "fill") if "y" in aes else "y"
        if "NORM" in kw: 
            aes[count] = "1e2*(..count..)/tapply(..count..,..PANEL..,sum)[..PANEL..]"
            labs[count] = "'fraction (%)'"
        else:
            aes[count] = "..count.."
            labs[count] = "'count'"
        args = "aes(%s), " % ", ".join(k + "=" + v for k,v in aes.items())
        if "INT" in kw: args += "binwidth=1, origin=-0.5, "
        if count == "y":   rcode.append("+ geom_histogram(%sposition='dodge')" % args)
        else:            rcode.append("+ geom_bin2d(%s)" % args)
        if "LOGZ" in kw: 
            rcode.append("+ scale_%s_continuous(trans='log')" % count)
            labs[count] = "''" # hide LOGZ label, unreadable anyways.
    
    matchrule(["X"],     setcol("x"))
    matchrule(["COLOR"], setcol("colour"))
    matchrule(["SHAPE"], setcol("shape"))
    matchrule(["FILL"],  setcol("fill"))
    matchrule(["LABEL"], setcol("label"))
    matchrule(["Y", "HIST"], setcol("y")) # only for 2D HIST, where multiple y make no sense
    matchrule(["Y"], lambda _,__,col: "y" in labs or labs.update(y="labels[1,'%s']" % rname[col]))
    
    if PRINTTHEME and "colour" in aes: del aes["colour"]
    rcode.append("ggplot(data, aes(%s))" % ", ".join(k + "=" + v for k,v in aes.items()))
    
    matchrule(["X", "HIST"], hist)
    
    rcode.append("+ labs(%s)"            % ", ".join(k + "=" + v for k,v in labs.items()))
            
    matchrule(["Y", "BAR"],   lambda _,__,y: rcode.append("+ geom_bar(aes(y=%s), stat='identity', position='dodge')" % rname[y]))
    # legend for shape uses wrong draw_key here
    matchrule(["Y", "BOX"],   lambda _,__,y: rcode.append("+ geom_boxplot(aes(y=%s), outlier.shape=NULL)" % rname[y]))
    matchrule(["Y", "VIOLIN"],lambda _,__,y: rcode.append("+ geom_violin (aes(y=%s))" % rname[y]))
    matchrule(["Y", "LINE"],  lambda _,__,y: rcode.append("+ geom_line   (aes(y=%s))" % rname[y]))
    matchrule(["Y", "LINEDASH"],  lambda _,__,y: rcode.append("+ geom_line   (aes(y=%s), linetype='dotted')" % rname[y]))
    matchrule(["Y", "JITTER"],lambda _,__,y: rcode.append("+ geom_jitter (aes(y=%s))" % rname[y]))
    matchrule(["Y", "POINT"], lambda _,__,y: rcode.append("+ geom_point  (aes(y=%s))" % rname[y]))
    matchrule(["Y", "POINTDODGE"], lambda _,__,y: rcode.append("+ geom_point  (aes(y=%s), position=position_dodge(width=0.8))" % rname[y]))
    matchrule(["Y", "TEXT"],  lambda _,__,y: rcode.append("+ geom_text   (aes(y=%s))" % rname[y]))
    
    matchrule(["FACET"],  lambda _,__,col: rcode.append("+ facet_wrap(~%s)"  % rname[col]))
    matchrule(["XFACET"], setcol("xf"))
    matchrule(["YFACET"], setcol("yf"))
    if "xf" in aes or "yf" in aes: rcode.append("+ facet_grid(%s~%s)" % 
        (aes["yf"] if "yf" in aes else ".", aes["xf"] if "xf" in aes else "."))
    
    matchrule(["Y", "LOG"], lambda _,__,___: rcode.append("+ scale_y_log10()"))
    matchrule(["X", "LOG"], lambda _,__,___: rcode.append("+ scale_x_log10()"))
    matchrule(["Y", "LOG2"], lambda _,__,___: rcode.append("+ scale_y_continuous(trans='log2')"))
    matchrule(["X", "LOG2"], lambda _,__,___: rcode.append("+ scale_x_continuous(trans='log2')"))
    matchrule(["Y", "FINE"], lambda _,__,___: rcode.append("+ scale_y_continuous(minor_breaks = seq(10, 1000, 10), breaks = seq(100, 1000, 100))"))
    matchrule(["X", "FINE"], lambda _,__,___: rcode.append("+ scale_x_continuous(minor_breaks = seq(10, 1000, 10), breaks = seq(100, 1000, 100))"))
    matchrule(["X", "LOGY"], lambda _,__,___: rcode.append("+ scale_y_log10()"))
    matchrule(["X", "ZERO"], lambda kw,__,___: rcode.append("+ expand_limits(x=%i)" % (1 if "LOG" in kw else 0)))
    matchrule(["Y", "ZERO"], lambda kw,__,___: rcode.append("+ expand_limits(y=%i)" % (1 if "LOG" in kw else 0)))
    matchrule(["X", "FLIP"], lambda _,__,___: rcode.append("+ coord_flip()"))
    matchrule(["X", "REVERSE"], lambda _,__,___: rcode.append("+ scale_x_reverse()"))
    matchrule(["Y", "REVERSE"], lambda _,__,___: rcode.append("+ scale_y_reverse()"))
    
    if PRINTTHEME:
        rcode.append("+ scale_fill_grey(start = 0, end = .8)")
        rcode.append("+ theme_Publication()")
        rcode.append("+ theme(legend.position = c(0.25,-0.3))")
        
    matchrule(["X", "VERT"], lambda _,__,___: rcode.append("+ theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5))"))
    
    rcode.append("+ scale_shape_manual(values=1:10)")
    matchrule(["ONSIDE"], lambda _,__,___: rcode.append("+ theme(legend.position = 'right')"))
    matchrule(["BELOW"], lambda _,__,___: rcode.append("+ theme(legend.position = 'bottom')"))
    
    if "''" in labs.values(): rcode.append("+ guides(%s)" % ", ".join('%s=FALSE' % k for k, v in labs.items() if v == "''"))
    rcode.append(")")
    rcode.append("ggsave(labels[1, 'outfilename'], plot, width=%f, height=%f)" % (w, h)) 
            
    with NamedTemporaryFile(delete=not manual) as t:
        conn.execute("ATTACH DATABASE ? AS forR;", (t.name,))
        conn.execute("CREATE TABLE forR.data(%s);" % 
                     ", ".join(rname[c] for c in cols))
        conn.execute("INSERT INTO forR.data %s" % sql)
        # avoid escape hell
        conn.execute("CREATE TABLE forR.labels(outfilename, %s);" % 
                     ", ".join(rname[c] for c in cols))
        conn.execute("INSERT INTO forR.labels VALUES(?,%s)" % ",".join(["?"]*len(cols)),
                     [outfile]+[lbl for _, lbl, _ in paras])
        conn.commit()
        r = "\n".join(r.replace("%s", t.name) for r in rcode).encode()
        print(r.decode())
        if not manual:
            p = subprocess.Popen(["Rscript", "-"], stdin=subprocess.PIPE)
            p.communicate(r)
