// -*- c-basic-offset: 4; tab-width: 8; indent-tabs-mode: t -*-        
#include "config.h"
#include <random>
#include <algorithm>
#include <sstream>
#include <strstream>
#include <iostream>
#include <string.h>
#include <math.h>
#include "network.h"
#include "pipe.h"
#include "eventlist.h"
#include "logfile.h"
#include "loggers.h"
#include "clock.h"
#include "ndp.h"
#include "compositequeue.h"
#include "topology.h"
#include "fat_tree_topology.h"

// Simulation params

#include "main.h"

uint32_t RTT = 1; // this is per link delay in us; identical RTT microseconds = 0.02 ms
#define DEFAULT_QUEUE_SIZE 8

string ntoa(double n);
string itoa(uint64_t n);

int KiB = 1024;
int MiB = KiB*KiB;
int flowsize_dist[] = {
    10*KiB, 10*KiB, 10*KiB,
    20*KiB,
    30*KiB, 30*KiB,
    50*KiB, 50*KiB,
    80*KiB, 80*KiB, 80*KiB,
    200*KiB,
    1*MiB, 1*MiB,
    2*MiB, 2*MiB,
    5*MiB, 5*MiB,
    10*MiB, 10*MiB,
};

 std::mt19937 rng;

class QueueTracer : public QueueLogger {
public:
    QueueTracer(string tag, double rate) {
        this->tag = tag;
        this->rate = rate;
    }
    void logQueue(Queue& queue, QueueEvent ev, Packet& pkt) {
        if (ev == PKT_ENQUEUE || ev == PKT_SERVICE) return;
        if (drand() > rate) return; 
        cout << tag << " at " << timeAsSec(queue.eventlist().now()) << " type " << ev << endl;
    }
private:
    string tag;
    double rate;
};

#include <fstream>
class TopologyFromFile : public Topology {
public:
    typedef unsigned __int128 bitset_t;
    TopologyFromFile(const char* filename, int concentration, 
            int n_paths, int queuesize, EventList* eventlist) {
        p = concentration;
        this->queuesize = queuesize;
        this->n_paths = n_paths;

        // 1. Read topology
        fstream topo(filename, ios_base::in);
        int n_v, n_e;
        int added_e = 0;
        topo >> n_v >> n_e;
        graph.resize(n_v);
        string ws; getline(topo, ws);
        for (int i = 0; i < n_v; i++) {
            string es;
            getline(topo, es);
            stringstream edges(es, ios_base::in);
            while (true) {
                int e;
                edges >> e;
                if (edges.eof()) break;
                graph[i].push_back(e-1);
                added_e++;
            }
        }
        assert(n_e*2 == added_e);

        // 2. set up links
        QueueLogger* src_log = new QueueTracer("Queue src", 1);
        QueueLogger* sink_log = new QueueTracer("Queue sink", 1);
        QueueLogger* net_log = new QueueTracer("Queue net", 0.01);
        inter_pipes.resize(n_v);
        inter_queues.resize(n_v);
        for (int s = 0; s < graph.size(); s++) {
            for (int t : graph[s]) {
                inter_pipes[s].push_back(new Pipe(timeFromUs(RTT), *eventlist));
                inter_queues[s].push_back(alloc_queue(eventlist, net_log));
            }
        }

        // 3. set up endpoints
        int edge_radix = n_v > 0 ? graph[0].size() : 0;
        for (int i = 0; i < n_v; i++) {
            if (graph[i].size() != edge_radix) break;
            for (int h = 0; h < p; h++) {
                end_in_pipes.push_back(new Pipe(timeFromUs(RTT), *eventlist));
                end_in_queues.push_back(alloc_queue(eventlist, sink_log));
                end_out_pipes.push_back(new Pipe(timeFromUs(RTT), *eventlist));
                end_out_queues.push_back(alloc_src_queue(eventlist, src_log));
            }
        }

        // 4. Compute MIN routing tables (ECMP capable)
        routing_table = make_min_routing(1);
    }

    virtual void make_ecmp_path(int s, int t, Route* routeout) {
        using param_type = typename std::uniform_int_distribution<int>::param_type;
        std::uniform_int_distribution<int> dist;
        bitset_t one = 1;
        int ttl = 10;
        while (s != t) {
            bitset_t next = (*routing_table)[s][t];
            int port;
            int ecmp_path = 0;
            for (port = 0; port < graph[s].size(); port++)
                if (next & (one << port)) ecmp_path++;
            ecmp_path = dist(rng, param_type{1, ecmp_path});
            for (port = 0; port < graph[s].size(); port++) {
                if (next & (one << port)) ecmp_path--;
                if (ecmp_path == 0) break;
            }
            routeout->push_back(inter_queues[s][port]);
            routeout->push_back(inter_pipes[s][port]);
            s = graph[s][port];
            assert(--ttl > 0);
        }
    }

    virtual vector<const Route*>* get_paths(int src,int dest) {
        vector<const Route*>* paths = new vector<const Route*>();
        int s = src/p, t = dest/p;
        std::uniform_int_distribution<> dist(0, graph.size()-1);

        // direct ECMP (path 0)
        Route* routeout = new Route();
        routeout->push_back(end_out_queues[src]);
        routeout->push_back(end_out_pipes[src]);
        make_ecmp_path(s, t, routeout);
        routeout->push_back(end_in_queues[dest]);
        routeout->push_back(end_in_pipes[dest]);
        paths->push_back(routeout);

        // VAL paths
        for (int i = 1; i < n_paths; i++) {
            Route* routeout = new Route();
            routeout->push_back(end_out_queues[src]);
            routeout->push_back(end_out_pipes[src]);

            int k = dist(rng);
            make_ecmp_path(s, k, routeout);
            make_ecmp_path(k, t, routeout);

            routeout->push_back(end_in_queues[dest]);
            routeout->push_back(end_in_pipes[dest]);
            paths->push_back(routeout);
        }
        return paths;
    }

    virtual vector<int>* get_neighbours(int src) {
        return &graph[src];  
    }

    virtual int no_of_nodes() const { return end_in_pipes.size(); }

private:
    Queue* alloc_src_queue(EventList* eventlist, QueueLogger* logger) {
        // host TX. NDP assumes exactly this class. Big buffer.
        return new PriorityQueue(speedFromMbps((uint64_t)HOST_NIC), memFromPkt(FEEDER_BUFFER), *eventlist, logger);
    }
    Queue* alloc_queue(EventList* eventlist, QueueLogger* logger) {
        // queue type that actually truncates... not sure wha the differences are
        return new CompositeQueue(speedFromMbps((uint64_t)HOST_NIC), memFromPkt(queuesize), *eventlist, logger);
    }

    vector<vector<bitset_t>>* make_min_routing(double rho) {
        int n_v = graph.size();
        int n_e = 0;
        for (auto& oute : graph) n_e += oute.size();

        // make a sparsified mask, by shuffling n_e ones and zeros, then reshaping.
        vector<bool> maskbits(n_e, false);
        for (int i = 0; i < (n_e*rho); i++) maskbits[i] = true;
        shuffle(maskbits.begin(), maskbits.end(), rng);
        vector<vector<bool>> mask(n_v);
        int j = 0;
        for (int i = 0; i < n_v; i++) for(auto e : graph[i]) mask[i].push_back(maskbits[j++]); 
        
        // bitset_t used as set of next hop ports.
        auto routing_table = new vector<vector<bitset_t>>(n_v, vector<bitset_t>(n_v, 0));
        vector<bitset_t> seen(n_v, 0);
        vector<bitset_t> next(n_v, 0);
        bitset_t one = 1;
        for (int s = 0; s < n_v; s++) {
            // one BFS graph flood per source, recording routes to all dest's
            for (int k = 0; k < n_v; k++) seen[k] = next[k] = 0;
            for (int port = 0; port < graph[s].size(); port++) {
                if (mask[s][port] == false) continue;
                next[graph[s][port]] = one << port;
            }
            int todo = n_v;
            while (true) {
                int newnow = 0;
                for (int t = 0; t < n_v; t++) {
                    if (next[t] && !seen[t]) {
                        (*routing_table)[s][t] = next[t];
                        newnow++; todo--;
                    }
                    seen[t] |= next[t];
                }
                if (newnow == 0 || todo == 0) break;
                for (int i = 0; i < n_v; i++) {
                    if (seen[i]) {
                        for (int port = 0; port < graph[i].size(); port++) {
                            if (mask[i][port] == false) continue;
                            next[graph[i][port]] |= seen[i];
                        }
                    }
                }
            }
            //for (int t = 0; t < n_v; t++)
            //    cerr << (*routing_table)[s][t] << " ";
            //cerr << endl;
        }
        return routing_table;
    }

    int p;
    int queuesize;
    int n_paths;
    vector<vector<int>> graph;
    vector<vector<bitset_t>>* routing_table;
    vector<vector<Pipe*>> inter_pipes;
    vector<vector<Queue*>> inter_queues;
    vector<Pipe*> end_in_pipes;
    vector<Pipe*> end_out_pipes;
    vector<Queue*> end_in_queues;
    vector<Queue*> end_out_queues;
};

class WorkloadFromFile {
public:
    map<int,vector<int>> connections;
    int no_of_conns() {
        return _n;
    }

    WorkloadFromFile(const char* filename) {
        _n = 0;
        fstream work(filename, ios_base::in);
        int n_v, n_e;
        work >> n_v >> n_e;
        string ws; getline(work, ws);
        for (int s = 0; s < n_v; s++) {
            string es;
            getline(work, es);
            stringstream edges(es, ios_base::in);
            while (true) {
                int t;
                edges >> t;
                if (edges.eof()) break;
                connections[s].push_back(t-1);
                _n++;
            }
        }
        //assert(n_e*2 == _n);
    }
private:
   int _n;
};


EventList eventlist;

Logfile* lg;

void exit_error(char* progr) {
    cerr << "Usage " << progr << " <sth>" << endl;
    exit(1);
}

int main(int argc, char **argv) {
    Packet::set_packet_size(9000);
    eventlist.setEndtime(timeFromSec(3));
    Clock c(timeFromSec(5 / 1000.), eventlist);
    int no_of_flows = 20, cwnd = 15, queuesize = 8;
    double lambda = 200;
    double pullrate = 1;
    int p = 0;
    int n_paths = 10;
    double q = 10;
    RouteStrategy route_strategy = FLOWLET;
    stringstream filename(ios_base::out);
    stringstream topofile(ios_base::out);
    stringstream workfile(ios_base::out);

    int i = 1;
    filename << "logout.dat";

    while (i<argc) {
        if (!strcmp(argv[i],"-o")){
            filename.str(std::string());
            filename << argv[i+1];
            i++;
        } else if (!strcmp(argv[i],"-topo")){
            topofile.str(std::string());
            topofile << argv[i+1];
            i++;
        } else if (!strcmp(argv[i],"-p")){
            p = atoi(argv[i+1]);
            i++;
        // use compatible names here
        } else if (!strcmp(argv[i],"-n_layers")){
            n_paths = atoi(argv[i+1]);
            i++;
        } else if (!strcmp(argv[i],"-rho")){
            q = atof(argv[i+1]);
            i++;
        } else if (!strcmp(argv[i],"-pullrate")){
            pullrate = atof(argv[i+1]);
            i++;
        } else if (!strcmp(argv[i],"-work")){
            workfile.str(std::string());
            workfile << argv[i+1];
            i++;
        } else if (!strcmp(argv[i],"-flows")){
            no_of_flows = atoi(argv[i+1]);
            i++;
        } else if (!strcmp(argv[i],"-lambda")){
            lambda = atoi(argv[i+1]);
            i++;
        } else if (!strcmp(argv[i],"-cwnd")){
            cwnd = atoi(argv[i+1]);
            i++;
        } else if (!strcmp(argv[i],"-qs")){
            queuesize = atoi(argv[i+1]);
            i++;
        } else if (!strcmp(argv[i],"-FLOWLET")){
            route_strategy = FLOWLET;
        } else if (!strcmp(argv[i],"-SCATTER_PERMUTE")){
            route_strategy = SCATTER_PERMUTE;
        } else {
            exit_error(argv[0]);
        }
        i++;
    }
    srand(42);
    rng.seed(42);
    cerr << "Using flows " << no_of_flows << endl;
      
    // prepare the loggers
    cerr << "Logging to " << filename.str() << endl;
    //Logfile 
    Logfile logfile(filename.str(), eventlist);

    int cnt_con = 0;
    int no_of_nodes = 0;
    int no_of_conns = 0;

    lg = &logfile;

    logfile.setStartTime(timeFromSec(0));
    //NdpSinkLoggerSampling sinkLogger = NdpSinkLoggerSampling(timeFromMs(10), eventlist);
    //logfile.addLogger(sinkLogger);
    //NdpTrafficLogger traffic_logger = NdpTrafficLogger();
    //logfile.addLogger(traffic_logger);
    NdpSrc* ndpSrc;
    NdpSink* ndpSnk;

    Route* routeout, *routein;

    NdpRtxTimerScanner ndpRtxScanner(timeFromMs(10), eventlist);
   
    int dest;

    Topology* top;

    if (topofile.str().length() >= 8 && 0==topofile.str().compare(topofile.str().length()-8, 8, "baseline")) {
        no_of_nodes = p*p*p*2;
        top = new FatTreeTopology(no_of_nodes, memFromPkt(queuesize), &logfile, &eventlist, nullptr, COMPOSITE,  0);
        cerr << "Made baseline FT with " << no_of_nodes << " nodes" << endl;
    } else if (topofile.str().length() >= 10 && 0==topofile.str().compare(topofile.str().length()-10, 10, "baseline2x")) {
        no_of_nodes = p*p*p/8*2*2;
        top = new FatTreeTopology(no_of_nodes, memFromPkt(queuesize), &logfile, &eventlist, nullptr, COMPOSITE,  0, 2 /* oversubscribe */);
        cerr << "Made oversubscribed FT with " << no_of_nodes << " nodes" << endl;
    } else {
        top = new TopologyFromFile(topofile.str().c_str(), p, n_paths, queuesize, &eventlist);
        cerr << "Loaded topo " << topofile.str() << endl;
    }

    no_of_nodes = top->no_of_nodes();
    cerr << "actual nodes " << no_of_nodes << endl;

    vector<vector<vector<const Route*>*>> net_paths(no_of_nodes, 
        vector<vector<const Route*>*>(no_of_nodes, nullptr));

    // Permutation connections
    WorkloadFromFile* conns = new WorkloadFromFile(workfile.str().c_str());
    no_of_conns = conns->no_of_conns();
    
    cerr << "Running with " << no_of_conns << " connections" << endl;

    vector <NdpSrc*> ndp_srcs;
    // initialize all sources/sinks
    NdpSrc::setMinRTO(50000); //increase RTO to avoid spurious retransmits
    NdpSrc::setRouteStrategy(route_strategy);
    NdpSink::setRouteStrategy(route_strategy);
    map<int,  NdpPullPacer*> pacers;

    int connID = 0;
    for (auto it : conns->connections){
        int src = it.first;
        auto destinations = it.second;
      
        for (unsigned int dst_id = 0;dst_id<destinations.size();dst_id++){
            connID++;
            dest = destinations.at(dst_id);
            if (!net_paths[src][dest]) {
                vector<const Route*>* paths = top->get_paths(src,dest);
                net_paths[src][dest] = paths;
            }
            if (!net_paths[dest][src]) {
                vector<const Route*>* paths = top->get_paths(dest,src);
                net_paths[dest][src] = paths;
            }

            NdpPullPacer*& pacer = pacers[dest];
            if (!pacer) pacer =new NdpPullPacer(eventlist, pullrate, cwnd);

            for (int flow = 0; flow < no_of_flows; flow++) {
                ndpSrc = new NdpSrc(NULL, NULL, eventlist);
                ndpSrc->setCwnd(cwnd*Packet::data_packet_size());
                ndpSrc->setHybQ(int64_t(q*Packet::data_packet_size()));
                ndp_srcs.push_back(ndpSrc);
                ndpSnk = new NdpSink(pacer);
                  
                ndpSrc->setName("ndp_" + ntoa(src) + "_" + ntoa(dest)+"("+ntoa(flow)+")");
                logfile.writeName(*ndpSrc);
                  
                ndpSnk->setName("ndp_sink_" + ntoa(src) + "_" + ntoa(dest)+ "("+ntoa(flow)+")");
                logfile.writeName(*ndpSnk);
                  
                ndpRtxScanner.registerNdp(*ndpSrc);
                  
                int choice = 0;
                routeout = new Route(*(net_paths[src][dest]->at(choice)));
                routeout->push_back(ndpSnk);
                  
                routein = new Route(*(net_paths[dest][src]->at(choice)));
                routein->push_back(ndpSrc);

                // even flows in first half, odd in second
                double extrastarttime = ((double)no_of_flows)/lambda/2 * (drand() + (flow%2));
                int flowsize = flowsize_dist[flow%(sizeof(flowsize_dist)/sizeof(flowsize_dist[0]))];
                  
                ndpSrc->set_flowsize(flowsize);
                ndpSrc->connect(*routeout, *routein, *ndpSnk, timeFromSec(extrastarttime));
                  
                if (route_strategy != SINGLE_PATH) {
                    ndpSrc->set_paths(net_paths[src][dest]);
                    ndpSnk->set_paths(net_paths[dest][src]);
                }
          
                //sinkLogger.monitorSink(ndpSnk);
            }
        }
    }

    cerr << "Loaded " << connID << " connections in total\n";

    // Record the setup
    int pktsize = Packet::data_packet_size();
    logfile.write("# pktsize=" + ntoa(pktsize) + " bytes");
    logfile.write("# flows=" + ntoa(no_of_flows));
    logfile.write("# lambda=" + ntoa(lambda));
    logfile.write("# hostnicrate = " + ntoa(HOST_NIC) + " pkt/sec");
    logfile.write("# corelinkrate = " + ntoa(HOST_NIC*CORE_TO_HOST) + " pkt/sec");
    double rtt = timeAsSec(timeFromUs(RTT));
    logfile.write("# rtt =" + ntoa(rtt));

    // GO!
    while (eventlist.doNextEvent()) {
    }
    cerr << "Done" << endl;

    vector <NdpSrc*>::iterator src_i;
    for (src_i = ndp_srcs.begin(); src_i != ndp_srcs.end(); src_i++) {
        //cerr << "Src, sent: " << (*src_i)->_packets_sent << "[new: " << (*src_i)->_new_packets_sent << " rtx: " << (*src_i)->_rtx_packets_sent << "] nacks: " << (*src_i)->_nacks_received << " pulls: " << (*src_i)->_pulls_received << " paths: " << (*src_i)->_paths.size() << endl;
    }
    for (src_i = ndp_srcs.begin(); src_i != ndp_srcs.end(); src_i++) {
        //(*src_i)->print_stats();
    }
}

string ntoa(double n) {
    stringstream s;
    s << n;
    return s.str();
}

string itoa(uint64_t n) {
    stringstream s;
    s << n;
    return s.str();
}
