#include "ECNTagger.h"
#include "inet/networklayer/ipv4/IPv4Datagram.h"
#include "inet/linklayer/ethernet/EtherFrame.h"
#include "inet/common/queue/SchedulerBase.h"

namespace inet {

Define_Module(ECNTagger);

simsignal_t ECNTagger::tagPkSignal = registerSignal("tagPk");

void ECNTagger::initialize()
{
    SchedulerBase::initialize();
    numGates = gateSize("in");
    for (int i = 0; i < numGates; ++i)
    {
        cGate *inGate = gate("in", i);
        cGate *connectedGate = inGate->getPathStartGate();
        if (!connectedGate)
            throw cRuntimeError("ECNTagger in gate is not connected");
        IQueueAccess *inModule = dynamic_cast<IQueueAccess*>(connectedGate->getOwnerModule());
        if (!inModule)
            throw cRuntimeError("ECNTagger in gate should be connected a simple module implementing IQueueAccess");
        queues.push_back(inModule);
        queueSet.insert(inModule);
    }

    // ECN dropper
    cGate *outGate = gate("dropperEcnOut");
    cGate *connectedGate = outGate->getPathEndGate();
    if (!connectedGate)
        throw cRuntimeError("ECNTagger dropperEcnOut gate is not connected");
    dropperEcn = dynamic_cast<AlgorithmicDropperBase*>(connectedGate->getOwnerModule());
    if (!dropperEcn)
        throw cRuntimeError("ECNTagger dropperEcnOut gate should be connected by a "
                            "simple module subclassing AlgorithmicDropperBase");
}

void ECNTagger::handleMessage(cMessage *msg)
{
    cPacket *packet = check_and_cast<cPacket*>(msg);
    if (shouldTag(packet)) {
        EV << "ECNTagger: ECN CE tagging packet\n";
        tagPacket(packet);
    } else {
        EV << "ECNTagger: Sending packet\n";
    }
    SchedulerBase::handleMessage(msg);
}


bool ECNTagger::shouldTag(cPacket *packet)
{
    return dropperEcn->shouldDrop(packet);
}

void ECNTagger::tagPacket(cPacket *packet)
{
    EthernetIIFrame *frame = dynamic_cast<EthernetIIFrame*>(packet);
    if (!frame) {
        return;
    }

    cPacket *encapPk = frame->decapsulate();
    IPv4Datagram *datagram = dynamic_cast<IPv4Datagram*>(encapPk);
    if (!datagram) {
        frame->encapsulate(encapPk);
        return;
    }

    datagram->setExplicitCongestionNotification(ECN_CE);
    frame->encapsulate(datagram);
    emit(tagPkSignal, packet);
}

int ECNTagger::getLength() const
{
    int len = 0;
    for (std::set<IQueueAccess*>::const_iterator it = queueSet.begin(); it != queueSet.end(); ++it)
        len += (*it)->getLength();
    return len;
}

int ECNTagger::getByteLength() const
{
    int len = 0;
    for (std::set<IQueueAccess*>::const_iterator it = queueSet.begin(); it != queueSet.end(); ++it)
        len += (*it)->getByteLength();
    return len;
}

}
