#include "ECMPRoutingTable.h"

namespace inet {
using namespace tcp;

void ECMPRoutingTable::initialize(int stage) {
    IPv4RoutingTable::initialize(stage);
    if (stage == INITSTAGE_LOCAL) {
        routerid = 0;
        if (par("useRouterId")) {
            const char* name = this->getParentModule()->getName();
            int len = strlen(name);
            if (len > 4) {
                const char* tail = name + (len-4);
                uint32 id;
                memcpy(&id, tail, 4);
                this->routerid = id;
                EV << "using id " << id << " for router "  << name << " tail " << tail << "\n";
            }
        }
    }
}

IPv4Route *ECMPRoutingTable::findBestMatchingRoute(IPv4Datagram *datagram) const
{
    Enter_Method("findBestMatchingRoute(IPv4Datagram *)");
    
    // Search in reverse in a ringbuffer -- so if we see the same destination
    // multiple times in a row, we have a hit immediatly.
    RouteVector const* bestRoutes = nullptr;
    for (size_t i = (cacheptr-1) % routeCache.size(); i != cacheptr; i = (i-1) % routeCache.size()) {
        if (routeCache[i].first == datagram->getDestAddress()) {
            bestRoutes = &(routeCache[i].second);
            break;
        }
    }
    
    if (bestRoutes == nullptr) {
        routeCache[cacheptr] = std::make_pair( 
            datagram->getDestAddress(),
            findBestMatchingRoutes(datagram->getDestAddress())
        );
        bestRoutes = &(routeCache[cacheptr].second);
        cacheptr = (cacheptr+1) % routeCache.size();
    }
    
    for (auto& r : *bestRoutes) {
        EV << "ecmp route " << r->info() << "\n";
    }

    if (bestRoutes->empty()) {
        return NULL;
    } else {
        RouteVector::const_iterator ecmpRoute = bestRoutes->begin();
        // TODO: this is not RFC 2992 hash-threshold
        std::advance(ecmpRoute, ecmpHash(datagram) % bestRoutes->size());
        EV << "ecmp chose " << (**ecmpRoute).info() << "\n";

        return *ecmpRoute;
    }
}

uint32 ECMPRoutingTable::ecmpHash(IPv4Datagram *datagram) const
{
    Enter_Method("ecmpHash(IPv4Datagram *)");

    int srcAddr = datagram->getSrcAddress().getInt();
    int dstAddr = datagram->getDestAddress().getInt();

    // I use the FNV-1a hash algorithm
    // The parameters are taken from http://isthe.com/chongo/tech/comp/fnv/#FNV-param
    uint32 fnv_prime = 16777619;
    uint32 hash = 2166136261;

    hash = (hash ^ routerid) * fnv_prime; // is 0 if disabled
    
    hash = (hash ^ srcAddr) * fnv_prime;
    hash = (hash ^ dstAddr) * fnv_prime;

    // For TCP also use dst/src port in the hash
    if (datagram->getTransportProtocol() == IP_PROT_TCP) {
        TCPSegment* tcpPacket = (TCPSegment*) datagram->getEncapsulatedPacket();
        hash = (hash ^ tcpPacket->getSrcPort()) * fnv_prime;
        hash = (hash ^ tcpPacket->getDestPort()) * fnv_prime;
    }

    return hash;
}

}
