#include "MPTCP.h"

#include "inet/common/INETUtils.h"
#include "inet/networklayer/common/IPSocket.h"
#include "inet/networklayer/contract/ipv4/IPv4ControlInfo.h"
#include "inet/networklayer/contract/ipv6/IPv6ControlInfo.h"
#include "inet/common/lifecycle/LifecycleOperation.h"
#include "inet/common/ModuleAccess.h"
#include "inet/common/lifecycle/NodeOperations.h"
#include "inet/common/lifecycle/NodeStatus.h"
#include "inet/transportlayer/tcp/TCPConnection.h"
#include "inet/transportlayer/tcp_common/TCPSegment.h"
#include "inet/transportlayer/contract/tcp/TCPCommand_m.h"

#ifdef WITH_IPv4
#include "inet/networklayer/ipv4/ICMPMessage_m.h"
#include "inet/networklayer/ipv4/IPv4Datagram.h"
#endif

#ifdef WITH_IPv6
#include "inet/networklayer/icmpv6/ICMPv6Message_m.h"
#endif

#include "inet/transportlayer/tcp/queues/TCPByteStreamRcvQueue.h"
#include "inet/transportlayer/tcp/queues/TCPByteStreamSendQueue.h"
#include "inet/transportlayer/tcp/queues/TCPMsgBasedRcvQueue.h"
#include "inet/transportlayer/tcp/queues/TCPMsgBasedSendQueue.h"
#include "inet/transportlayer/tcp/queues/TCPVirtualDataRcvQueue.h"
#include "inet/transportlayer/tcp/queues/TCPVirtualDataSendQueue.h"

#include "inet/transportlayer/mptcp/MPTCPSendQueue.h"
#include "inet/transportlayer/mptcp/queues/MPTCPVirtualDataSendQueue.h"

namespace inet {

Define_Module(MPTCP);

Register_Class(MPTCPUncoupled);
Register_Class(MPTCPCoupled);
Register_Class(MPTCPECNCoupled);
Register_Class(DCMPTCP);

MPTCP::MPTCP() {
}

MPTCP::~MPTCP()
{
    while (!mptcpAppConnMap.empty()) {
        MptcpAppConnMap::iterator i = mptcpAppConnMap.begin();
        delete i->second;
        mptcpAppConnMap.erase(i);
    }
    tcpAppConnMap.clear();
}

MPTCPConnection *MPTCP::createMultipathConnection(int appGateIndex, int connId)
{
    MPTCPConnection *conn = new MPTCPConnection(this, appGateIndex, connId);
    TCP::AppConnKey key;
    key.appGateIndex = appGateIndex;
    key.connId = connId;
    tcpAppConnMap[key] = conn->subflows.at(0);
    mptcpAppConnMap[key] = conn;
    return conn;
}

void MPTCP::handleMessage(cMessage *msg)
{
    if (!isOperational)
    {
        if (msg->isSelfMessage())
            throw cRuntimeError("Model error: self msg '%s' received when isOperational is false", msg->getName());
        EV << "MPTCP is turned off, dropping '" << msg->getName() << "' message\n";
        delete msg;
    }
    else if (msg->isSelfMessage())
    {
        MPTCPSubflow *subflow = (MPTCPSubflow *) msg->getContextPointer();
        bool ret = subflow->processTimer(msg);
        if (!ret)
            removeConnection(subflow);
    }
    else if (msg->arrivedOn("ipIn") || msg->arrivedOn("ipv6In"))
    {
        if (false
#ifdef WITH_IPv4
                || dynamic_cast<ICMPMessage *>(msg)
#endif
#ifdef WITH_IPv6
                || dynamic_cast<ICMPv6Message *>(msg)
#endif
            )
        {
            EV << "ICMP error received -- discarding\n"; // FIXME can ICMP packets really make it up to TCP???
            delete msg;
        }
        else
        {
            // must be a TCPSegment
            TCPSegment *tcpseg = check_and_cast<TCPSegment *>(msg);

            if (tcpseg->getEceBit()) {
                EV << "Segment has CE bit set\n";
            }

            // get src/dest addresses
            L3Address srcAddr, destAddr;

            if (dynamic_cast<IPv4ControlInfo *>(tcpseg->getControlInfo()) != NULL)
            {
                IPv4ControlInfo *controlInfo = (IPv4ControlInfo *)tcpseg->removeControlInfo();
                srcAddr = controlInfo->getSrcAddr();
                destAddr = controlInfo->getDestAddr();
                tcpseg->setEceBit(tcpseg->getEceBit() || controlInfo->getExplicitCongestionNotification() == ECN_CE);
                tcpseg->setCwrBit(true);
                delete controlInfo;
            }
            else if (dynamic_cast<IPv6ControlInfo *>(tcpseg->getControlInfo()) != NULL)
            {
                IPv6ControlInfo *controlInfo = (IPv6ControlInfo *)tcpseg->removeControlInfo();
                srcAddr = controlInfo->getSrcAddr();
                destAddr = controlInfo->getDestAddr();
                delete controlInfo;
            }
            else
            {
                error("(%s)%s arrived without control info", tcpseg->getClassName(), tcpseg->getName());
            }

            // process segment
            MPTCPSubflow *subflow = (MPTCPSubflow*) findConnForSegment(tcpseg, srcAddr, destAddr);
            if (subflow)
            {
                bool ret = subflow->processTCPSegment(tcpseg, srcAddr, destAddr);
                if (!ret)
                    removeConnection(subflow);
            }
            else
            {
                segmentArrivalWhileClosed(tcpseg, srcAddr, destAddr);
            }
        }
    }
    else // must be from app
    {
        TCPCommand *controlInfo = check_and_cast<TCPCommand *>(msg->getControlInfo());
        int appGateIndex = msg->getArrivalGate()->getIndex();
        int connId = controlInfo->getConnId();

        MPTCPConnection *conn = findMultipathConnForId(appGateIndex, connId);
        //MPTCPSubflow *subflow = (MPTCPSubflow*) findConnForApp(appGateIndex, connId);

        if (!conn)//!subflow)
        {
            conn = createMultipathConnection(appGateIndex, connId);

            EV << "MPTCP connection created for " << msg << "\n";
        }

        bool ret;
        //if (!conn)
        //    ret = subflow->processAppCommand(msg);
        //else
            ret = conn->processAppCommand(msg);

        //if (!ret)
        //    removeConnection(subflow);
    }

    // TODO: ev
    //if (ev.isGUI())
    //    updateDisplayString();
}

TCPOption *MPTCP::getMPCapableOption(TCPSegment *tcpseg) {
    for (int i = 0; i < tcpseg->getHeaderOptionArraySize(); i++) {
        TCPOptionPtr *mp_capable = &tcpseg->getHeaderOption(i);
        if ((*mp_capable)->getKind() == TCPOPTION_MPCAPABLE)
            return *mp_capable;
    }
    return NULL;
}

TCPSendQueue* MPTCP::createSendQueue(TCPDataTransferMode transferModeP)
{
    return new MPTCPSendQueueAdapter();
}

MPTCPConnection *MPTCP::findMultipathConnForSegment(TCPSegment *tcpseg, L3Address srcAddr, L3Address destAddr)
{
    //TCPOption *mpCapableOption = getMPCapableOption(tcpseg);
    //int connId = mpCapableOption->getValues(2);
    MPTCPConnection *conn = NULL;//findMultipathConnForId(connId);

    if (!conn) {
        MPTCPSubflow *subflow = (MPTCPSubflow*) findConnForSegment(tcpseg, srcAddr, destAddr);
        conn = subflow->mptcpConn;
    }
    
    return conn;
}

MPTCPConnection *MPTCP::findMultipathConnForId(int appGateIndex, int connId)
{
    TCP::AppConnKey key;
    key.appGateIndex = appGateIndex;
    key.connId = connId;
    MptcpAppConnMap::iterator i = mptcpAppConnMap.find(key);
    return i == mptcpAppConnMap.end() ? NULL : i->second;
}

void MPTCP::removeMultipathConnection(MPTCPConnection *conn)
{

}

MPTCPConnection::MPTCPConnection()
{
}

MPTCPConnection::MPTCPConnection(MPTCP *_mptcp, int _appGateIndex, int _connId)
{
    mptcpMain = _mptcp;
    int numSubflows = mptcpMain->par("mptcpSubflows").longValue();
    for (int i = 0; i < numSubflows; i++) {
        MPTCPSubflow *subflow = new MPTCPSubflow(mptcpMain, this, _appGateIndex, _connId);
        addSubflow(subflow);
        subflow->createAlgorithm(NULL);
    }
    sendQueue = new MPTCPVirtualDataSendQueue(this);
    subflows.at(0)->primary = true;
    // TODO: ev
    mptcpKey = 42; //ev.getUniqueNumber();
    nextSubflow = 0;
}

MPTCPConnection::~MPTCPConnection()
{
    for (std::vector<MPTCPSubflow *>::iterator it = subflows.begin() ; it != subflows.end(); ++it)
        delete (*it);
    subflows.clear();
    delete sendQueue;
}

void MPTCPConnection::addSubflow(MPTCPSubflow *subflow) {
    subflows.push_back(subflow);
}

bool MPTCPConnection::processTimer(cMessage *msg)
{
    return getNextSubflow()->processTimer(msg);
}

bool MPTCPConnection::processTCPSegment(TCPSegment *tcpSeg, L3Address srcAddr, L3Address destAddr)
{
    return getNextSubflow()->processTCPSegment(tcpSeg, srcAddr, destAddr);
}

bool MPTCPConnection::processAppCommand(cMessage *msg)
{
    TCPCommand *tcpCmd = check_and_cast<TCPCommand *>(msg->getControlInfo());
    int appGateIndex = msg->getArrivalGate()->getIndex();
    int connId = tcpCmd->getConnId();
    TCPEventCode event = TCPConnection::preanalyseAppCommandEvent(msg->getKind());

    switch (event) {
    case TCP_E_SEND:
        return processTcpSend(msg, tcpCmd);
    case TCP_E_OPEN_PASSIVE:
        return subflows.at(0)->processAppCommand(msg);
    case TCP_E_CLOSE:
        delete msg;
        return true;//subflows.at(0)->processAppCommand(msg);
    default:
        break;
    }

    for (MPTCPSubflow *subflow : subflows) {
        cMessage *msg_copy = msg->dup();
        msg_copy->setControlInfo(tcpCmd->dup());
        subflow->processAppCommand(msg_copy);
    }
    delete msg;
    return true;
}

bool MPTCPConnection::processTcpSend(cMessage *msg, TCPCommand *tcpCmd)
{
    cPacket *pk_subflow = PK(msg)->dup();
    TCPCommand *tcpCmdCpy = tcpCmd->dup();
    sendQueue->enqueueAppData(PK(msg));
    pk_subflow->setByteLength(0);
    for (MPTCPSubflow *subflow : subflows) {
        cPacket *cpy = pk_subflow->dup();
        cpy->setControlInfo(tcpCmdCpy->dup());
        subflow->processAppCommand(cpy);
        delete cpy;
    }
    delete tcpCmdCpy;
    delete pk_subflow;
    return true;
}

MPTCPSubflow *MPTCPConnection::getNextSubflow()
{
    uint ret = nextSubflow;
    nextSubflow = (nextSubflow + 1) % subflows.size();
    return subflows.at(ret);
}

TCPConnection *MPTCPSubflow::create(TCP *mod, int appGateIndex, int connId)
{
    return new MPTCPSubflow(mod, NULL, appGateIndex, connId);
}

MPTCPSubflow::MPTCPSubflow(TCP *_tcp, MPTCPConnection *_conn, int _appGateIndex, int _connId)
: TCPConnection(_tcp, _appGateIndex, _connId)
{
    mptcpConn = _conn;
}

MPTCPSubflow::~MPTCPSubflow()
{
    // Yes, this is a hack.
    sendQueue = new MPTCPSendQueueAdapter();
}

TCPSegment MPTCPSubflow::writeHeaderOptions(TCPSegment *tcpseg)
{
    TCPConnection::writeHeaderOptions(tcpseg);
    TCPOptionUnknown *option = new TCPOptionUnknown();
    option->setKind(TCPOPTION_MPCAPABLE);
    option->setLength(5);
    option->setBytesArraySize(3);
    option->setBytes(2, this->mptcpConn->mptcpKey);
    tcpseg->addHeaderOption(option);
    return *tcpseg;
}

TCPConnection *MPTCPSubflow::cloneListeningConnection()
{
    MPTCPSubflow *subflow = (MPTCPSubflow*) TCPConnection::cloneListeningConnection();
    subflow->mptcpConn = mptcpConn;
    check_and_cast<MPTCPSendQueueAdapter *>(subflow->sendQueue)->setMPTCPConnection(mptcpConn);
    mptcpConn->addSubflow(subflow);
    return subflow;
}

void MPTCPSubflow::initConnection(TCPOpenCommand *openCmd)
{
    TCPConnection::initConnection(openCmd);
    ASSERT(mptcpConn);
    ((MPTCPSendQueueAdapter *) sendQueue)->setMPTCPConnection(mptcpConn);
}

TCPAlgorithm *MPTCPSubflow::createAlgorithm(const char *tcpAlgorithmClass)
{
    if (!tcpAlgorithmClass || !tcpAlgorithmClass[0])
        tcpAlgorithmClass = mptcpConn->mptcpMain->par("mptcpAlgorithmClass");
    MPTCPAlgorithm *mptcpAlgorithm = check_and_cast<MPTCPAlgorithm *>(inet::utils::createOne(tcpAlgorithmClass ));
    mptcpAlgorithm->setConnection(this);
    mptcpAlgorithm->subflows = &mptcpConn->subflows;
    return mptcpAlgorithm;
}

void MPTCPSubflow::sendEstabIndicationToApp() {
    if (primary) {
        TCPConnection::sendEstabIndicationToApp();
    }
}

TCPRenoStateVariables *MPTCPSubflow::getState()
{
    return ((MPTCPAlgorithm*) tcpAlgorithm)->getState();
}

TCPRenoStateVariables *MPTCPAlgorithm::getState()
{
    return state;
}

void MPTCPAlgorithm::receivedDataAck(uint32 firstSeqAcked)
{
    TCPTahoeRenoFamily::receivedDataAck(firstSeqAcked);

    if (state->dupacks >= DUPTHRESH) // DUPTHRESH = 3
    {
        // Perform Fast Recovery: set cwnd to ssthresh (deflating the window).
        EV << "Fast Recovery: setting cwnd to ssthresh=" << state->ssthresh << "\n";
        state->snd_cwnd = state->ssthresh;

        if (cwndVector)
            cwndVector->record(state->snd_cwnd);
    }
    else
    {
        // Perform slow start and congestion avoidance.
        if (state->snd_cwnd < state->ssthresh)
        {
            EV << "cwnd <= ssthresh: Slow Start: increasing cwnd by one SMSS bytes to ";
            state->snd_cwnd += state->snd_mss;
            if (cwndVector)
                cwndVector->record(state->snd_cwnd);

            EV << "cwnd=" << state->snd_cwnd << "\n";
        }
        else
        {
            // perform Congestion Avoidance (RFC 2581)
            congestionAvoidance(firstSeqAcked);
        }
    }
    sendData(false);
}

void MPTCPAlgorithm::congestionAvoidance(uint32 firstSeqAcked)
{
    uint32 bytes_acked = state->snd_una - firstSeqAcked;
    uint32 incr = inflateCwnd(bytes_acked);

    if (incr == 0)
        incr = 1;

    state->snd_cwnd += incr;

    if (cwndVector)
        cwndVector->record(state->snd_cwnd);

    EV << "cwnd > ssthresh: Congestion Avoidance: increasing cwnd linearly, to " << state->snd_cwnd << "\n";
}

uint32 MPTCPAlgorithm::inflateCwnd(uint32 bytes_acked)
{
    return state->snd_mss * state->snd_mss / state->snd_cwnd;
}

uint32 MPTCPUncoupled::inflateCwnd(uint32 bytes_acked)
{
    return state->snd_mss * state->snd_mss / state->snd_cwnd;
}

uint32 MPTCPCoupled::inflateCwnd(uint32 bytes_acked)
{
    uint cwd_total = computeTotalCwnd();
    uint32 alpha = computeAlpha(cwd_total);
    uint32 tcp_inc = state->snd_mss * state->snd_mss / state->snd_cwnd;
    uint32 mptcp_inc = alpha * bytes_acked * state->snd_mss / (ALPHA_SCALE * cwd_total);
    return std::min(mptcp_inc, tcp_inc);
}

uint32 MPTCPCoupled::computeAlpha(uint32 cwd_total)
{
    uint32 max_cwnd_rtt = 0;
    uint32 sum_cwnd_rtt = 0;
    for (MPTCPSubflow *subflow : *subflows) {
        TCPTahoeRenoFamilyStateVariables *state_i = subflow->getState();
        int rtt_i = state_i->srtt.inUnit(SIMTIME_MS);
        rtt_i = rtt_i?rtt_i:1; // If RTT is zero then we set it to one MS
        uint32 cwnd_rtt_i = state_i->snd_cwnd /(rtt_i * rtt_i);
        max_cwnd_rtt = std::max(max_cwnd_rtt, cwnd_rtt_i);
        cwd_total += cwnd_rtt_i * cwnd_rtt_i;
    }
    return cwd_total;
}
uint32 MPTCPCoupled::computeTotalCwnd()
{
    uint32 cwd_total = 0;
    for (MPTCPSubflow *subflow : *subflows) {
        cwd_total += subflow->getState()->snd_cwnd;
    }
    return cwd_total;
}

void MPTCPECNCoupled::receivedDataAck(uint32 firstSeqAcked)
{
    TCPTahoeRenoFamily::receivedDataAck(firstSeqAcked);

    if (hasEcnMark()) { // ECN bit set
        if (!ecnAcks) {
            // enter Fast Recovery
            recalculateSlowStartThreshold();
            // set cwnd to ssthresh plus 3 * SMSS.
            state->snd_cwnd = state->ssthresh + 3 * state->snd_mss;
            EV << "Congestion Avoidance: got ecn bit, entering fast recovery, updating cwnd=" << state->snd_cwnd << "\n";
            ecnWnd = state->snd_max - state->snd_una;
        } else if (ecnAcks * state->snd_mss < ecnWnd) {
            EV << "Congestion Avoidance: ecnAcks=" << ecnAcks << " < " << ecnWnd / state->snd_mss << ", staying in fast recovery\n";
        } else {
            state->snd_cwnd = state->snd_mss;
            EV << "Congestion Avoidance: Too much congestion, updating cwnd to snd_mss=" << state->snd_mss << "\n";
        }
        ecnAcks++;
    } else {
        if (ecnAcks >= ECN_THRESH) {
            // Similar to fast recovery
            state->snd_cwnd = state->ssthresh;
            EV << "Congestion Avoidance: got ecn bit, in fast recovery, updating cwnd to ssthresh=" << state->ssthresh << "\n";
            ecnAcks = 0;
        } else {
            // Congestion avoidance
            congestionAvoidance(firstSeqAcked);
        }
    }
    sendData(false);
}

uint32 MPTCPECNCoupled::inflateCwnd(uint32 bytes_acked)
{
    double alpha = subflows->at(0)->mptcpConn->mptcpMain->par("alpha").doubleValue();
    return alpha * state->snd_mss * state->snd_mss / state->snd_cwnd;
}

DCMPTCP::DCMPTCP() {
    alpha = 1;
    ecnAcks = 0;
    nonEcnAcks = 0;
}

void DCMPTCP::receivedDataAck(uint32 firstSeqAcked)
{
    if (state->dctcp_ece) {
        TCPTahoeRenoFamily::receivedDataAck(firstSeqAcked);
        ecnAcks++;
        state->snd_cwnd *= 1 - alpha/2;
        if (cwndVector)
            cwndVector->record(state->snd_cwnd);
        sendData(false);
    } else {
        nonEcnAcks++;
        MPTCPAlgorithm::receivedDataAck(firstSeqAcked);
    }
}

void DCMPTCP::rttMeasurementComplete(simtime_t tSent, simtime_t tAcked)
{
    double ecnFrac = ecnAcks / (1.0 * ecnAcks + nonEcnAcks);
    /* alpha = (1-g) * alpha + g * F */
    alpha = (1 - DCTCP_G) * alpha + DCTCP_G * ecnFrac;
    ecnAcks = nonEcnAcks = 0;
    TCPBaseAlg::rttMeasurementComplete(tSent, tAcked);
}

}

