//
// Copyright (C) 2015 Erik Henriksson
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
//

#ifndef __INET_MPTCPMAIN_H
#define __INET_MPTCPMAIN_H

#include <map>
#include <set>

#include "inet/common/INETDefs.h"
#include "inet/common/lifecycle/ILifecycle.h"
#include "inet/networklayer/common/L3Address.h"
#include "inet/transportlayer/tcp/TCPConnection.h"
#include "inet/transportlayer/contract/tcp/TCPCommand_m.h"
#include "inet/transportlayer/tcp/TCP.h"
#include "inet/transportlayer/tcp/TCPAlgorithm.h"
#include "inet/transportlayer/tcp/flavours/TCPReno.h"
#include "inet/transportlayer/tcp/flavours/DCTCP.h"

namespace inet {
using namespace tcp;

class MPTCP;
class MPTCPAlgorithm;
class MPTCPConnection;
class MPTCPSubflow;
class MPTCPCoupled;
class MPTCPSendQueue;
class MPTCPSendQueueAdapter;

enum MPTCP_Option {
    MP_CAPABLE = 0,
    MP_JOIN = 1,
    DSS = 2,
    ADD_ADDR = 3,
    REMOVE_ADDR = 4,
    MP_PRIO = 5,
    MP_FAIL = 6,
    MP_FASTCLOSE = 7
};

enum MPTCP_Flags {
    CHECKSUM_REQUIRED = 0,
    HMAC_SHA1 = 7
};

// An RFC 6824 implementation of Multipath TCP

class INET_API MPTCP : public TCP
{

protected:
  typedef std::map<TCP::AppConnKey, MPTCPConnection*> MptcpAppConnMap;
  typedef std::map<SockPair, MPTCPConnection*> MptcpConnMap;

  MptcpAppConnMap mptcpAppConnMap;
  MptcpConnMap mptcpConnMap;

public:
    MPTCP();
    virtual ~MPTCP();
    virtual TCPSendQueue* createSendQueue(TCPDataTransferMode transferModeP);
protected:
    virtual MPTCPConnection *createMultipathConnection(int appGateIndex, int connId);
    virtual void handleMessage(cMessage *msg);
    virtual MPTCPConnection *findMultipathConnForSegment(TCPSegment *tcpseg, L3Address srcAddr, L3Address destAddr);
    virtual MPTCPConnection *findMultipathConnForId(int appGateIndex, int connId);
    virtual void removeMultipathConnection(MPTCPConnection *conn);
    virtual TCPOption *getMPCapableOption(TCPSegment *tcpseg);
};

class INET_API MPTCPConnection
{
public:
    MPTCP *mptcpMain;
    MPTCPSendQueue *sendQueue;
    std::vector<MPTCPSubflow*> subflows;
    uint mptcpKey;
    uint nextSubflow;
private:
public:
    MPTCPConnection();
    MPTCPConnection(MPTCP *mptcp, int appGateIndex, int connId);
    virtual ~MPTCPConnection();
    virtual void addSubflow(MPTCPSubflow *subflow);
    virtual bool processTimer(cMessage *msg);
    virtual bool processTCPSegment(TCPSegment *tcpSeg, L3Address srcAddr, L3Address destAddr);
    virtual bool processAppCommand(cMessage *msg);
    virtual bool processTcpSend(cMessage *msg, TCPCommand *tcpCmd);
    virtual MPTCPSubflow *getNextSubflow();
};

class INET_API MPTCPSubflow : public TCPConnection
{
public:
    MPTCPConnection* mptcpConn;
    bool primary = false;
public:
    MPTCPSubflow(TCP *_tcp, MPTCPConnection *_conn, int _appGateIndex, int _connId);
    virtual ~MPTCPSubflow();
    TCPConnection *create(TCP *mod, int appGateIndex, int connId);
    TCPRenoStateVariables *getState();
protected:
    virtual void initConnection(TCPOpenCommand *openCmd);
    virtual TCPAlgorithm *createAlgorithm(const char *tcpAlgorithmClass);
    TCPSegment writeHeaderOptions(TCPSegment *tcpseg);
    TCPConnection *cloneListeningConnection();
    virtual void sendEstabIndicationToApp();
friend MPTCPConnection;
};

class INET_API MPTCPAlgorithm : public TCPReno // TCPAlgorithm
{
public:
    std::vector<MPTCPSubflow*> *subflows;
public:
    MPTCPAlgorithm() {state = NULL; conn = NULL;}
    virtual ~MPTCPAlgorithm() {}
    TCPRenoStateVariables *getState();
protected:
    virtual void receivedDataAck(uint32 firstSeqAcked);
    virtual uint32 inflateCwnd(uint32 bytes_acked);
    virtual void congestionAvoidance(uint32 firstSeqAcked);
};

// RFC 2581 / RFC 5681 TCP Reno
class INET_API MPTCPUncoupled : public MPTCPAlgorithm
{
public:
    MPTCPUncoupled() {}
    virtual ~MPTCPUncoupled() {}
protected:
    virtual uint32 inflateCwnd(uint32 bytes_acked);
};

// RFC 6356 Coupled Congestion Control for Multipath Transport Protocols
class INET_API MPTCPCoupled : public MPTCPAlgorithm
{
public:
    MPTCPCoupled() {}
    virtual ~MPTCPCoupled() {}
    static const uint32 ALPHA_SCALE = 512;
protected:
    virtual uint32 inflateCwnd(uint32 bytes_acked);
    uint32 computeAlpha(uint32 cwd_total);
    uint32 computeTotalCwnd();
};

#define ECN_THRESH (10)

// Experimental ECN Coupled Congestion Control for MPTCP
class INET_API MPTCPECNCoupled : public MPTCPCoupled
{
public:
    MPTCPECNCoupled() {}
    virtual ~MPTCPECNCoupled() {}
protected:
    int ecnAcks = 0;
    uint32 ecnWnd = 0;

    virtual void receivedDataAck(uint32 firstSeqAcked);
    virtual uint32 inflateCwnd(uint32 bytes_acked);
    inline bool hasEcnMark() {return state->dctcp_ece;}
};

class INET_API DCMPTCP : public MPTCPCoupled
{
  protected:
    double alpha;
    int ecnAcks;
    int nonEcnAcks;
    virtual void rttMeasurementComplete(simtime_t tSent, simtime_t tAcked);
  public:
    /** Ctor */
    DCMPTCP();
    virtual void receivedDataAck(uint32 firstSeqAcked);
};

}

#endif
