#ifndef __INET_MPTCPSENDQUEUE_H
#define __INET_MPTCPSENDQUEUE_H

#include "inet/common/INETDefs.h"
#include "inet/transportlayer/mptcp/MPTCP.h"
#include "inet/transportlayer/tcp/TCPConnection.h"
#include "inet/transportlayer/tcp_common/TCPSegment.h"
#include "inet/transportlayer/tcp/TCPSendQueue.h"


#include <algorithm>

namespace inet {
using namespace tcp;

class INET_API MPTCPSendQueue : public cObject
{
  protected:
    MPTCPConnection *conn;

  public:
    MPTCPSendQueue()  {conn = NULL;}
    virtual ~MPTCPSendQueue() {}
    virtual void setConnection(MPTCPConnection *_conn)  {conn = _conn;}
    virtual void init(int subflowId, uint32 startSeq) = 0;
    virtual void enqueueAppData(cPacket *msg) = 0;

    /**
     * Returns the sequence number of the first byte stored in the buffer.
     */
    virtual uint32 getBufferStartSeq(int subflowId) = 0;

    /**
     * Returns the sequence number of the last byte stored in the buffer plus one.
     * (The first byte of the next send operation would get this sequence number.)
     */
    virtual uint32 getBufferEndSeq(int subflowId) = 0;

    /**
     * Utility function: returns how many bytes are available in the queue, from
     * (and including) the given sequence number.
     */
    inline ulong getBytesAvailable(int subflowId, uint32 fromSeq)
    {
        uint32 bufEndSeq = getBufferEndSeq(subflowId);
        return seqLess(fromSeq, bufEndSeq) ? bufEndSeq-fromSeq : 0;
    }

    /**
     * Called when the TCP wants to send or retransmit data, it constructs
     * a TCP segment which contains the data from the requested sequence
     * number range. The actually returned segment may contain less than
     * maxNumBytes bytes if the subclass wants to reproduce the original
     * segment boundaries when retransmitting.
     */
    virtual TCPSegment *createSegmentWithBytes(int subflowId, uint32 fromSeq, ulong maxNumBytes) = 0;

    /**
     * Tells the queue that bytes up to (but NOT including) seqNum have been
     * transmitted and ACKed, so they can be removed from the queue.
     */
    virtual void discardUpTo(int subflowId, uint32 seqNum) = 0;

    virtual void reserveUpTo(int subflowId, uint32 seqNum) = 0;

};

class INET_API MPTCPSendQueueAdapter : public TCPSendQueue
{
protected:
  MPTCPConnection *mptcpConn; // The MPTCPConnection
  int _subflowId;

public:
  MPTCPSendQueueAdapter() {mptcpConn = NULL; _subflowId = -1;}
  virtual ~MPTCPSendQueueAdapter() {}
  virtual void setConnection(TCPConnection *_conn)  {conn = _conn;}
  virtual void setMPTCPConnection(MPTCPConnection *_conn)  {mptcpConn = _conn;}
  virtual int subflowId()
  {
      if (_subflowId == -1) {
          std::vector<MPTCPSubflow*>::iterator it = std::find(mptcpConn->subflows.begin(), mptcpConn->subflows.end(), this->conn);
          ASSERT(it != mptcpConn->subflows.end());
          _subflowId = it - mptcpConn->subflows.begin();
      }
      return _subflowId;
  }
  virtual void init(uint32 startSeq) {mptcpConn->sendQueue->init(subflowId(), startSeq);}
  virtual void enqueueAppData(cPacket *msg) {}
  virtual uint32 getBufferStartSeq() {return mptcpConn->sendQueue->getBufferStartSeq(subflowId());}
  virtual uint32 getBufferEndSeq() {return mptcpConn->sendQueue->getBufferEndSeq(subflowId());}
  virtual TCPSegment *createSegmentWithBytes(uint32 fromSeq, ulong maxNumBytes)
  {
      return mptcpConn->sendQueue->createSegmentWithBytes(subflowId(), fromSeq, maxNumBytes);
  }
  virtual void discardUpTo(uint32 seqNum) {mptcpConn->sendQueue->discardUpTo(subflowId(), seqNum);}
  virtual void reserveUpTo(uint32 seqNum) {mptcpConn->sendQueue->reserveUpTo(subflowId(), seqNum);}
};

Register_Class(MPTCPSendQueueAdapter);

}

#endif
