//
// Copyright (C) 2004-2005 Andras Varga
// Copyright (C) 2009 Thomas Reschka
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
//

#include <algorithm>   // min,max
#include "inet/transportlayer/tcp/flavours/DCTCP.h"
#include "inet/transportlayer/tcp/flavours/TCPReno.h"
#include "inet/transportlayer/tcp/flavours/TCPBaseAlg.h"
#include "inet/transportlayer/tcp/TCP.h"

namespace inet {
using namespace tcp;

Register_Class(DCTCP);

DCTCP::DCTCP() {
    dctcp_alpha = 0;
    acked_bytes_ecn = 0;
    acked_bytes_total = 0;
}

void DCTCP::receivedDataAck(uint32 firstSeqAcked)
{
    // Adjust cwnd (TX side)
    uint32 old_cwnd = state->snd_cwnd;
    int64 mid_cwnd = -1;
    acked_bytes_total++;
    if (state->dctcp_ece) {
        EV << "ECN: DCTCP responding to ECE ACK\n";
        acked_bytes_ecn++;
        // Allowing less makes connections stuck with no space to send, no packets 
        // in flight to clear up the CE state and no RTO setup (which would normally 
        // save it,  but there is no loss here).
        if (state->snd_cwnd > state->snd_mss) {
            // HACK: only use multiples of mss here. Since all logic we use is packet-based,
            // having a cwnd that is not a multiple of mss causes small packets to be sent,
            // which mess up all the other logic.
            state->snd_cwnd = state->snd_mss * int(state->snd_cwnd / state->snd_mss * (1.0 - dctcp_alpha/2.0));
        } else {
            state->snd_cwnd = state->snd_mss;
        }
        if (cwndVector)
            cwndVector->record(state->snd_cwnd);
        mid_cwnd = state->snd_cwnd;
    }
    // Not sure if this reflects the spec, but better do all the complicated stuff in any case.
    TCPReno::receivedDataAck(firstSeqAcked);
    // TCPReno will increase the cwnd by one mss if all is fine. Compensate it here.
    // Otherwise there is no immediate feedback on CE (stop increasing cwnd), which
    // makes ECN useless for extreme burst loads.
    if (state->dctcp_ece) {
        if (state->snd_cwnd > state->snd_mss) 
            state->snd_cwnd = state->snd_cwnd - state->snd_mss;
    }
    
    uint32 end_cwnd = state->snd_cwnd;
    EV_INFO << "DCTCP: cwnd update " << conn->connId << " : " << old_cwnd << " " << mid_cwnd << " " << end_cwnd << "\n";
}

void DCTCP::receiveSeqChanged()
{
    // Update dctcp_ce state machine and generate ACKs (RX side)
    if (state->dctcp_ce == false) { // CE=0
        if (state->dctcp_ecn == true) {
            EV << "ECN: DCTCP responding to segment with ECN\n";
            state->dctcp_ce = true;
            state->ack_now = true; // Send immediate ACK with ECE=1
        }
    } else { // CE=1
        if (state->dctcp_ecn == false) {
            EV << "ECN: DCTCP_CE reset\n";
            state->dctcp_ce = false;
            state->ack_now = true; // Send immediate ACK with ECE=0
        }
    }
    // Send 1 ACK for every m packets with ECE=dctcp_ce.
    // Also sends the immediate ACKs if requested.
    TCPBaseAlg::receiveSeqChanged();
        
}

void DCTCP::rttMeasurementComplete(simtime_t tSent, simtime_t tAcked)
{
    dctcp_alpha = (1 - DCTCP_G) * dctcp_alpha + DCTCP_G * acked_bytes_ecn / acked_bytes_total;
    acked_bytes_ecn = acked_bytes_total = 0;
    TCPBaseAlg::rttMeasurementComplete(tSent, tAcked);
}

}
