//=========================================================================
//  CNULLMESSAGEPROT.CC - part of
//
//                  OMNeT++/OMNEST
//           Discrete System Simulation in C++
//
//  Author: Andras Varga, 2003
//          Dept. of Electrical and Computer Systems Engineering,
//          Monash University, Melbourne, Australia
//
//=========================================================================

/*--------------------------------------------------------------*
  Copyright (C) 2003-2015 Andras Varga
  Copyright (C) 2006-2015 OpenSim Ltd.

  This file is distributed WITHOUT ANY WARRANTY. See the file
  `license' for details on this and other legal matters.
*--------------------------------------------------------------*/

#include "omnetpp/cmessage.h"
#include "omnetpp/cmodule.h"
#include "omnetpp/cgate.h"
#include "omnetpp/cenvir.h"
#include "omnetpp/cconfiguration.h"
#include "omnetpp/cparsimcomm.h"
#include "omnetpp/ccommbuffer.h"
#include "omnetpp/globals.h"
#include "omnetpp/cconfigoption.h"
#include "omnetpp/regmacros.h"
#include "omnetpp/cchannel.h"
#include "omnetpp/cfutureeventset.h"
#include "cffnullmessageprot.h"
#include "../src/sim/parsim/clinkdelaylookahead.h"
#include "../src/sim/parsim/cparsimpartition.h"
#include "../src/sim/parsim/messagetags.h"
#include "../src/sim/parsim/cplaceholdermod.h"
#include "../src/sim/parsim/cproxygate.h"

#include <cstdlib>

namespace omnetpp {

Register_Class(cFFNullMessageProtocol);
Register_GlobalConfigOption(CFGID_PARSIM_NULLMESSAGEPROTOCOL_LOOKAHEAD_CLASS, "parsim-nullmessageprotocol-lookahead-class", CFG_STRING, "cLinkDelayLookahead", "When `cNullMessageProtocol` is selected as parsim synchronization class: specifies the C++ class that calculates lookahead. The class should subclass from `cNMPLookahead`.");
Register_GlobalConfigOption(CFGID_PARSIM_NULLMESSAGEPROTOCOL_LAZINESS, "parsim-nullmessageprotocol-laziness", CFG_DOUBLE, "0.5", "When `cNullMessageProtocol` is selected as parsim synchronization class: specifies the laziness of sending null messages. Values in the range `[0,1)` are accepted. Laziness=0 causes null messages to be sent out immediately as a new EOT is learned, which may result in excessive null message traffic.");
extern cConfigOption *CFGID_PARSIM_DEBUG;  // registered in cparsimpartition.cc

cFFNullMessageProtocol::cFFNullMessageProtocol() : cParsimProtocolBase()
{
    numSeg = 0;
    segInfo = nullptr;

    debug = getEnvir()->getConfig()->getAsBool(CFGID_PARSIM_DEBUG);
    std::string lookhClass = getEnvir()->getConfig()->getAsString(CFGID_PARSIM_NULLMESSAGEPROTOCOL_LOOKAHEAD_CLASS);
    lookaheadcalc = dynamic_cast<cNMPLookahead *>(createOne(lookhClass.c_str()));
    if (!lookaheadcalc) \
        throw cRuntimeError("Class \"%s\" is not subclassed from cNMPLookahead", lookhClass.c_str());
    laziness = getEnvir()->getConfig()->getAsDouble(CFGID_PARSIM_NULLMESSAGEPROTOCOL_LAZINESS);
}

cFFNullMessageProtocol::~cFFNullMessageProtocol()
{
    delete lookaheadcalc;

    // segInfo[*].eitEvent/eotEvent messages already deleted with getFES()->clear()
    delete[] segInfo;
}

void cFFNullMessageProtocol::setContext(cSimulation *sim, cParsimPartition *seg, cParsimCommunications *co)
{
    cParsimProtocolBase::setContext(sim, seg, co);
    lookaheadcalc->setContext(sim, seg, co);
}

void cFFNullMessageProtocol::startRun()
{
    EV_INFO << "starting FF Null Message Protocol...\n";

    delete[] segInfo;

    numSeg = comm->getNumPartitions();
    segInfo = new PartitionInfo[numSeg];
    int myProcId = comm->getProcId();
    // if it is seeded, good; otherwise make at least sure all segments have different sequences.
    std::srand(std::rand() + myProcId);

    char buf[30];
    int i;

    // temporarily initialize everything to zero.
    for (i = 0; i < numSeg; i++) {
        segInfo[i].eotEvent = nullptr;
        segInfo[i].eitEvent = nullptr;
        segInfo[i].lastEotSent = 0.0;

        segInfo[i].time = 0;
        segInfo[i].messageBalance = 0;
        segInfo[i].nextEventGuess = 0.0;
    }

    active = true;
    messageBalance = 0;
    segInfo[myProcId].time = 0;
    acktime = 0;

    lookahead = 0;

    // Note boot sequence: first we have to schedule all "resend-EOT" events,
    // so that the simulation will start by sending out null messages --
    // otherwise we'd end up sitting blocked on an EIT event forever!

    // create "resend-EOT" events and schedule them to zero (1st thing to do)
    EV << "  scheduling 'resend-EOT' events...\n";
    for (i = 0; i < numSeg; i++) {
        if (i != myProcId) {
            sprintf(buf, "resendEOT-%d", i);
            cMessage *eotMsg = new cMessage(buf, MK_PARSIM_RESENDEOT);
            eotMsg->setContextPointer((void *)(long)i);  // khmm...
            segInfo[i].eotEvent = eotMsg;
            rescheduleEvent(eotMsg, 0.0);
        }
    }

    // create EIT events and schedule them to zero (null msgs will bump them)
    EV << "  scheduling 'EIT' events...\n";
    for (i = 0; i < numSeg; i++) {
        if (i != myProcId) {
            sprintf(buf, "EIT-%d", i);
            cMessage *eitMsg = new cMessage(buf, MK_PARSIM_EIT);
            segInfo[i].eitEvent = eitMsg;
            rescheduleEvent(eitMsg, 0.0);
        }
    }

    // start lookahead calculator too
    lookaheadcalc->startRun();

    EV << "  setup done.\n";
}

void cFFNullMessageProtocol::endRun()
{
    lookaheadcalc->endRun();
}

void cFFNullMessageProtocol::processOutgoingMessage(cMessage *msg, int destProcId, int destModuleId, int destGateId, void *data)
{
    messageBalance++;
    active = true;
    // calculate lookahead
    simtime_t lookahead = lookaheadcalc->getCurrentLookahead(msg, destProcId, data);
    simtime_t eot = sim->getSimTime() + lookahead;
    if (eot < segInfo[destProcId].lastEotSent)
        throw cRuntimeError("cNullMessageProtocol error: attempt to decrease EOT");

    // send a null message only if EOT is better than last time
    bool sendNull = (eot > segInfo[destProcId].lastEotSent);

    // send message
    cCommBuffer *buffer = comm->createCommBuffer();
    if (sendNull) {
        // update "resend-EOT" timer
        segInfo[destProcId].lastEotSent = eot;
        simtime_t eotResendTime = sim->getSimTime() + lookahead*laziness;
        rescheduleEvent(segInfo[destProcId].eotEvent, eotResendTime);

        {if (debug) EV << "piggybacking null msg on `" << msg->getName() << "' to " << destProcId << ", lookahead=" << lookahead << ", EOT=" << eot << "; next resend at " << eotResendTime << "\n";}

        // send cMessage with piggybacked null message
        buffer->pack(eot);
        buffer->pack(destModuleId);
        buffer->pack(destGateId);
        buffer->packObject(msg);
        comm->send(buffer, TAG_CMESSAGE_WITH_NULLMESSAGE, destProcId);
    }
    else
    {
        {if (debug) EV << "sending `" << msg->getName() << "' to " << destProcId << "\n";}

        // send cMessage
        buffer->pack(destModuleId);
        buffer->pack(destGateId);
        buffer->packObject(msg);
        comm->send(buffer, TAG_CMESSAGE, destProcId);
    }
    comm->recycleCommBuffer(buffer);
}

void cFFNullMessageProtocol::processReceivedBuffer(cCommBuffer *buffer, int tag, int sourceProcId)
{
    int destModuleId;
    int destGateId;
    cMessage *msg;
    simtime_t eit;

    switch (tag) {
        case TAG_CMESSAGE_WITH_NULLMESSAGE:
            buffer->unpack(eit);
            processReceivedEIT(sourceProcId, eit);
            buffer->unpack(destModuleId);
            buffer->unpack(destGateId);
            msg = (cMessage *)buffer->unpackObject();
            processReceivedMessage(msg, destModuleId, destGateId, sourceProcId);
            messageBalance--;
            active = true;
            break;

        case TAG_CMESSAGE:
            buffer->unpack(destModuleId);
            buffer->unpack(destGateId);
            msg = (cMessage *)buffer->unpackObject();
            processReceivedMessage(msg, destModuleId, destGateId, sourceProcId);
            messageBalance--;
            active = true;
            break;

        case TAG_NULLMESSAGE:
            buffer->unpack(eit);
            buffer->unpack(segInfo[sourceProcId].time);
            buffer->unpack(segInfo[sourceProcId].messageBalance);
            buffer->unpack(segInfo[sourceProcId].nextEventGuess);
            processReceivedEIT(sourceProcId, eit);
            break;

        default:
            partition->processReceivedBuffer(buffer, tag, sourceProcId);
            break;
    }
    buffer->assertBufferEmpty();
}

void cFFNullMessageProtocol::processReceivedEIT(int sourceProcId, simtime_t eit)
{
    cMessage *eitMsg = segInfo[sourceProcId].eitEvent;

    {if (debug) EV << "null msg received from " << sourceProcId << ", EIT=" << eit << ", rescheduling EIT event\n";}

    // sanity check
    ASSERT(eit > eitMsg->getArrivalTime());

    // reschedule it to the EIT just received
    rescheduleEvent(eitMsg, eit);
}

cEvent *cFFNullMessageProtocol::takeNextEvent()
{
    // our EIT and resendEOT messages are always scheduled, so the FES can
    // only be empty if there are no other partitions at all -- "no events" then
    // means we're finished.
    if (sim->getFES()->isEmpty())
        return nullptr;

    // we could do a receiveNonblocking() call here to look at our mailbox,
    // but for performance reasons we don't -- it's enough to read it
    // (receiveBlocking()) when we're stuck on an EIT. Or should we do it
    // every 1000 events or so? If MPI receive buffer fills up it can cause
    // deadlock.
    // receiveNonblocking();

    cEvent *event;
    while (true) {
        event = sim->getFES()->peekFirst();
        cMessage *msg = event->isMessage() ? static_cast<cMessage *>(event) : nullptr;
        if (msg && msg->getKind() == MK_PARSIM_RESENDEOT) {
            // send null messages if window closed for a partition
            int procId = (long)msg->getContextPointer();  // khmm...
            sendNullMessage(procId, event->getArrivalTime());
        }
        else if (msg && msg->getKind() == MK_PARSIM_EIT) {
            // wait until it gets out of the way (i.e. we get a higher EIT)
            {if (debug) EV << "blocking on EIT event `" << event->getName() << "'\n";}
            tryFastForward();
            if (!receiveBlocking())
                return nullptr;
        }
        else {
            // just a normal event -- go ahead with it
            break;
        }
    }

    // remove event from FES and return it
    cEvent *tmp = sim->getFES()->removeFirst();
    ASSERT(tmp == event);
    active = true;
    return event;
}

bool cFFNullMessageProtocol::tryFastForward() {
    int myProcId = comm->getProcId();
    cEvent* event = sim->getFES()->peekFirst();
    cMessage *msg = event->isMessage() ? static_cast<cMessage *>(event) : nullptr;
    simtime_t horizon = 0;
    if (msg && msg->getKind() == MK_PARSIM_EIT) { // should always be the case
        horizon = event->getArrivalTime();
    } else {
        return false;
    }

    // The fact that we got here means we are now idle. If we were idle before,
    // we just check if there is termination; otherwise we try to cause a snapshot.

    // first, check time and termination.
    int64_t maxtime = 0;
    simtime_t nextev = segInfo[0].nextEventGuess;
    int64_t inflight = 0;
    bool snapshot_ok = !active; // no snapshot possible if we are active.
    for (int i = 0; i < numSeg; i++)
        if (segInfo[i].time > maxtime) maxtime = segInfo[i].time; 
    for (int i = 0; i < numSeg; i++) {
        if (segInfo[i].time == maxtime) {
            inflight += segInfo[i].messageBalance;
            if (nextev > segInfo[i].nextEventGuess) nextev = segInfo[i].nextEventGuess;
        } else {
            snapshot_ok = false;
        }
    }

    if ( ! snapshot_ok) { // this is the normal case, we are not converged yet
        if (active) { // we want to be the last one to idle!
            segInfo[myProcId].time = (maxtime & ~0xFFL) + 0x100 + myProcId;
            active = false;
        } else { // we are idle already, so we reply to the greatest
            if (segInfo[myProcId].time == maxtime) return false; // we are waiting for others
            segInfo[myProcId].time = maxtime;
        }
        // no we take a snapshot for the greatest (may be ourself)
        segInfo[myProcId].messageBalance = messageBalance;   
        // this does not need to be fast, if we got here there is no work to
        // do either way.
        std::vector<cEvent*> tempstore;
        cEvent* cur = sim->getFES()->peekFirst();
        while (cur && cur->isMessage() && (
                ((cMessage*) cur)->getKind() == MK_PARSIM_RESENDEOT ||  
                ((cMessage*) cur)->getKind() == MK_PARSIM_EIT)) {
            tempstore.push_back(sim->getFES()->removeFirst());
            cur = sim->getFES()->peekFirst();
        }
        // cur is the next "real" event
        segInfo[myProcId].nextEventGuess = cur->getArrivalTime();
        // now reinsert whatever we removed
        for (cEvent* ev : tempstore)
             sim->getFES()->insert(ev);

        // we responded and wait for the next round.
        return false;
    } else { // we got a fully valid global snapshot.
        if (acktime == maxtime) return false;
        acktime = maxtime;
        EV_INFO << "+++ valid snapshot " << nextev << " inflight  " << inflight << " time " << maxtime << " horizon " << horizon << "\n";
        if (inflight != 0)
            return false; // the snapshot is valid, but the state is not.
        // the snapshot might be outdated -- ignore it, then.
        if (nextev <= horizon)
            return false;
    
        // Sent new EOT to unblock others
        EV_INFO << "+++ fast forward  from " << " to " << nextev - lookahead / laziness << " at " << horizon << "\n";
        for (int i = 0; i < numSeg; i++) {
            if (i != myProcId) {
                sendNullMessage(i, nextev - lookahead / laziness); // make sure the peers still block to free us.
            }
        }

        return true;
    }
}

void cFFNullMessageProtocol::putBackEvent(cEvent *event)
{
    throw cRuntimeError("cNullMessageProtocol: \"Run Until Event/Module\" functionality "
                        "cannot be used with this scheduler (putBackEvent() not implemented)");
}

void cFFNullMessageProtocol::sendNullMessage(int procId, simtime_t now)
{
    if (lookahead == 0) {
        // force equal lookahead (and esp. all-to-all connections)
        lookahead = 1;
        for (int i = 0; i < numSeg; i++) 
            if (i != comm->getProcId() && lookaheadcalc->getCurrentLookahead(i) < lookahead)
                lookahead = lookaheadcalc->getCurrentLookahead(i);
    }

    // calculate EOT and sending of next null message
    simtime_t eot = now + lookahead;

    if (eot < segInfo[procId].lastEotSent)
        throw cRuntimeError("cNullMessageProtocol error: attempt to decrease EOT");
    segInfo[procId].lastEotSent = eot;

    // calculate time of next null message sending, and schedule "resend-EOT" event
    simtime_t eotResendTime = now + lookahead*laziness;
    rescheduleEvent(segInfo[procId].eotEvent, eotResendTime);

    {if (debug) EV << "sending null msg to " << procId << ", lookahead=" << lookahead << ", EOT=" << eot << "; next resend at " << eotResendTime << "\n";}

    // send out null message
    int myProcId = comm->getProcId();
    cCommBuffer *buffer = comm->createCommBuffer();
    buffer->pack(eot);
    buffer->pack(segInfo[myProcId].time);
    buffer->pack(segInfo[myProcId].messageBalance);
    buffer->pack(segInfo[myProcId].nextEventGuess);
    comm->send(buffer, TAG_NULLMESSAGE, procId);
    comm->recycleCommBuffer(buffer);
}

void cFFNullMessageProtocol::rescheduleEvent(cMessage *msg, simtime_t t)
{
    sim->getFES()->remove(msg);  // also works if the event is not currently scheduled
    msg->setArrivalTime(t);
    sim->getFES()->insert(msg);
}

}  // namespace omnetpp

