#!/usr/bin/env python3
import random
import numpy as np
from collections import defaultdict

edgetype = np.uint64

def read_graph(filename):
    f = open(filename)

    firstline = f.readline().split(" ")
    edges = int(firstline[1])
    vertices = int(firstline[0])
    
    graph = np.zeros((vertices, vertices), dtype=edgetype)
    
    # Reproduce whatever run_inet does when setting up the net.
    # Ports 0--(p-1) are for the p endpoints. we add this offset later.
    # A duplex link is added for each edge in the file iff src node (row) < dst node (col).
    # Port numbers are used in order. Which means we have to keep counts for the remote ends.
    # See also extract-topo.awk.
    #
    # Node ids are 1-based, but we use 0-base for the matrix.
    # We count internal ports 0-based and set 2**i for an edge on port i.
    # 
    # Note that we force the graph do be undirected due to duplex links everywhere,
    # but we have to handle it as a directed graph afterwards, since routes (port numbers!)
    # can differ in both directions.
    
    port = defaultdict(lambda: edgetype(0))
    vertex = 0
    one =  edgetype(1)
    for line in f:
        s = line.split(" ")
        vertex += 1
        for e in s:
            if e.isspace(): continue
            dest = int(e)
            if vertex < dest:
                graph[vertex-1, dest-1] |= one << port[vertex]
                port[vertex] += one
                graph[dest-1, vertex-1] |= one << port[dest]
                port[dest] += one                
                
    return graph


def make_shortest(g, args, count = 1):
    def left(a, b):
        out = np.array(a)
        out[b == 0] = 0
        return out
    def union(a):
        return np.bitwise_or.reduce(a)
    
    def mmm(outer, inner, a, b):
        out = np.zeros_like(a)
        for i in range(a.shape[0]):
            for j in range(a.shape[1]):
                out[i,j] = outer(inner(a[i,:], b[:,j]))
        return out
    
    def parallel_mmm(outer, inner, a, b):
        def partial_mmm(indices):
            out = np.zeros_like(a)
            for i in indices:
                if i >= a.shape[0]: continue
                for j in range(a.shape[1]):
                    out[i,j] = outer(inner(a[i,:], b[:,j]))
            return out
        chunk = (a.shape[0]//simplepmap.cpu_count())+1
        res = simplepmap.pmap(partial_mmm, [range(i*chunk, (i+1)*chunk) for i in range(simplepmap.cpu_count())])
        out = np.zeros_like(a)
        for r in res:
            out += r;
        return out
    
    try:
        import simplepmap
        mmm = parallel_mmm
    except:
        pass # sequential is also fine
    
    out = np.zeros(g.shape, dtype=np.uint64)
    
    acc = g
    
    for i in range(args.maxlen):
        newshortest = (out == 0) & (acc != 0)
        out[newshortest] = acc[newshortest]
        acc = mmm(union, left, acc, g)

    # encode_ecmp will randomly select on set of routes each, so we can just
    # return the full set for each layer
    return [out]*int(count)

def make_sparse(g, args, density = 0.5, count = 1):
    def sparsify(g, amount):
        g = np.array(g)
        h = g > 0
        total = int(h.sum())
        remaining = int(total*amount)
        values = np.array([1.0]*remaining + [0.0]*(total-remaining))
        random.shuffle(values)
        values *= g[h]
        g[h] = values
        return g
    out = [make_shortest(sparsify(g, float(density)), args, 1) for _ in  range(int(count))]
    return sum(out, [])

# We assign one /24 subnet per node. Within that, one /30 per endpoint link.
# For each link between nodes, we use a /30 of a disjoint pool, that is not 
# routable from the rest of the network. These adresses are not explicitly used 
# at all, as we route by interface.
# The adresses within a link do not really matter, as we route on subnet level.
# For each node, routing between the endpoints is automatic by netmask, and we
# need one table entry per node(-subnet) (potentially with multiple paths).
# We can separate layers either by ecmp or by using the same structure with
# different IP prefixes, so we can have shortest ath ECMP per layer on multiple 
# non-shortest path layers.
# We will hit lots of scalability problems with the stock IPv4NetworkConfigurator...
def encode_node(node, layer):
    return "%d.%d.%d.0" % (10, (layer << 5) + (node+1)//2**8, (node+1) % 2**8)
def encode_host(node, port, offset, layer):
    return "%d.%d.%d.%d" % (10, (layer << 5) + (node+1)//2**8, (node+1) % 2**8, port*4+offset)

def encode_addresses(f, g, args):
    for net in args.subnets:
        for node in range(g.shape[0]):
            for host in range(args.p):
                print('<interface hosts="Node%04d"         names="eth%d" netmask="255.255.255.252" address="%s" />' % (node+1, host, encode_host(node, host, 2, net)), file=f)
                print('<interface hosts="Endpoint_%04d_%02d" names="eth0" netmask="255.255.255.252" address="%s" />' % (node+1, host, encode_host(node, host, 1, net)), file=f)
                print('<route     hosts="Endpoint_%04d_%02d" destination="*" netmask="0.0.0.0" gateway="%s" interface="eth0" />' % (node+1, host,  encode_host(node, host, 2, net)), file=f)
        print("", file=f)
    global portmap
    portmap = {} # (host, port) -> (src ip, dst ip)
    addr = 0
    for src in range(g.shape[0]):
        for dst in range(g.shape[1]):
            if src < dst and g[src, dst] != 0:
                port1 = int(g[src, dst]).bit_length() - 1
                port2 = int(g[dst, src]).bit_length() - 1
                addr1 = "192.%d.%d.%d" % (168+(addr+1)//2**16, (addr+1)//2**8%2**8, (addr+1)%2**8)
                addr2 = "192.%d.%d.%d" % (168+(addr+2)//2**16, (addr+2)//2**8%2**8, (addr+2)%2**8)
                addr += 4
                print('<interface hosts="Node%04d" names="eth%d" netmask="255.255.255.252" address="%s" />' % (src+1, port1 + args.p, addr1), file=f)
                print('<interface hosts="Node%04d" names="eth%d" netmask="255.255.255.252" address="%s" />' % (dst+1, port2 + args.p, addr2), file=f)
                portmap[(src, port1)] = (addr1, addr2, "eth%d" % (port1 + args.p))
                portmap[(dst, port2)] = (addr2, addr1, "eth%d" % (port2 + args.p))

def encode_route(src, dst, port, layer):
    global portmap
    _, toaddr, interface = portmap[(src, port)]
    return ('<route hosts="Node%04d" destination="%s" netmask="255.255.255.0" gateway="%s" interface="%s" />' % 
            (src+1, encode_node(dst, layer), toaddr, interface))

def encode_ecmp(f, ls, args, net = 0):
    for l in ls:
        for src in range(l.shape[0]):
            for dst in range(l.shape[1]):
                if src == dst: continue
                slots = args.limit
                options = [port for port, b in enumerate(reversed(format(l[src, dst], "b"))) if b == '1']
                # randomize ports in case we have to limit the number
                random.shuffle(options)
                options = options[0:min(len(options), args.limit)]
                for port in options:
                    print(encode_route(src, dst, port, net), file=f)
                    
def encode_layered(f, ls, args):
    assert len(ls) == len(args.subnets)
    for l, net in zip(ls, args.subnets):
        encode_ecmp(f, [l], args, net)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(prog='make_routing.py')
    parser.add_argument('graph', help='Network graph file')
    parser.add_argument('-p', help='Number of endpoints per router.', type=int, default=1)
    parser.add_argument('--layer', help='Use these methods to generate more layers.', required=True)
    parser.add_argument('--encode', help='Use this packing function to encode nexthops into a routing table.', default='ecmp')
    parser.add_argument('--limit', help='Maximum number of routing table entries per route (ECMP)', type=int, default=1)
    parser.add_argument('--maxlen', help='Maximum allowed length of routes', type=int, default=5)
    parser.add_argument('--seed', help='Random seed', type=int, default=42)
    
    args = parser.parse_args()
    
    random.seed(args.seed)
    
    layers = {
        'shortest': make_shortest, # all shortest paths (takes multiplicity option)
        'sparse': make_sparse, # shortest paths in sprsified graph (takes sparsity and multiplicity options)
    }

    layerargs = [(name, args) for name, *args in 
        (opt.split('_') for opt in args.layer.split(','))]
    
    encoders = {
        'ecmp': encode_ecmp, # make one route for each path, possibly limited by --limit, for randomized ECMP
        'layered': encode_layered, # make exactly --limit routes per layer, potentially repeating routes
    }

    g = read_graph(args.graph)
    ls = [layers[l](g, args, *opts) for l, opts in layerargs]
    ls = sum(ls, [])
    encode_routes = encoders[args.encode]

    # address-based layers: only power of two, we use a mask between /8 and /11 to
    # group layers
    numnets = len(ls)
    assert numnets in [1, 2, 4, 8]
    args.subnets = [net for net in range(numnets)]
    
    with open("routing.xml", "w") as f:
        print("<config>", file=f)
        encode_addresses(f, g, args)
        print("", file=f)
        encode_routes(f, ls, args)
        print("</config>", file=f)
