#!/usr/bin/env python3
import random

def output_graph(graph, file=None):
    print(len(graph), int(sum(len(n) for n in graph) / 2), file=file)
    for node in graph:
       print( " ".join(str(e+1) for e in node) + " ", file=file)

def add_graph(g, h):
    assert len(g) == len(h)
    return [n1 + n2 for n1, n2 in zip(g,h)]

def sum_graphs(*gs):
    out = None
    for g in gs:
        out = add_graph(out, g) if out else g
    return out

def invert(perm):
    p = [None] * len(perm)
    for i in range(len(perm)): p[perm[i]] = i
    return p
       
def make_permutation(n):
    l = list(map(lambda i: [i], range(0, n)))
    random.shuffle(l)
    return l

def make_ring(n, stride):
    return list(map(lambda i: [i], range(n-stride, n))) + list(map(lambda i: [i], range(0, n-stride)))

def make_stencil(n, stride):
    return sum_graphs(make_ring(n, 1), make_ring(n, stride), make_ring(n, n-1), make_ring(n, n-stride))

def make_fourone(n, stride):
    return sum_graphs(make_stencil(n, stride), make_permutation(n))

def make_fourfour(n, stride):
    return sum_graphs(make_stencil(n, stride), map_random(make_stencil(n, stride)))

def map_random(graph):
    n = len(graph)
    p = list(range(0, n))
    random.shuffle(p)
    return [[p[t] for t in graph[p[s]]] for s in range(n)]

def map_blocked(graph, n, size):
    p = [[i for i in range(k, k+size) if i < n] for k in range(0, n+size, size)]
    random.shuffle(p)
    p = sum(p, [])
    pinv = invert(p)
    return [[pinv[t] for t in (graph[p[s]] if p[s] < len(graph) else [])] for s in range(n)]

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(prog='make_workload.py')
    parser.add_argument('--type', help='Type of workload (permutation, ...).', required=True)
    parser.add_argument('--map', help='Mapping of ranks to endpoints (linear, random, ...)', default="identity")
    parser.add_argument('-n', help='Number of routers.', type=int, required=True)
    parser.add_argument('-p', help='Number of endpoints per router.', type=int, default=1)
    parser.add_argument('--fraction', help='Fraction of endpoints participating in communication.', type=float, default=1.0)
    parser.add_argument('--stride', help='Offset between src and dst for ring (=k-th side diagonal for stencil).', type=int, default=1)
    parser.add_argument('--blocksize', help='Blocksize for blocked random mapping.', type=int, default=None)
    parser.add_argument('--seed', help='Random seed', type=int, default=42)
    
    args = parser.parse_args()
    
    random.seed(args.seed)
    
    types = {'permutation': lambda args: make_permutation(int(args.n*args.p*args.fraction)),
             'ring':        lambda args: make_ring(int(args.n*args.p*args.fraction), args.stride),
             'stencil':     lambda args: make_stencil(int(args.n*args.p*args.fraction), args.stride),
             'fourone':     lambda args: make_fourone(int(args.n*args.p*args.fraction), args.stride),
             'fourfour':    lambda args: make_fourfour(int(args.n*args.p*args.fraction), args.stride),
             }
    
    maps = {'identity': lambda args, graph: graph,
            'random':   lambda args, graph: map_random(graph),
            'blocked':  lambda args, graph: map_blocked(graph, args.n*args.p, args.blocksize or args.p),
            }
    
    graph = types[args.type](args)
    mapped = maps[args.map](args, graph)
    
    with open("workload.txt", "w") as f:
        output_graph(mapped, f)
